#!/usr/bin/env python3

# Copyright (C) 2020-2022 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/cactus_grid_ascii.py, cactus_grid_h5.py,
# cactus_grid_omni.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.cactus_grid` module provides functions to load grid function
in Cactus formats.

There are multiple classes defined in this module:

- :py:class`~.GridFunctionsDir` interfaces with :py:class:`~.SimDir` and
  organizes the grid functions by dimensionality. This is a dictionary-like
  object with keys the possible dimensions (e.g., ``x``, ``yz``, ``xyz``).
- :py:class`~.AllGridFunctions` takes all the files in SimDir and sort them
  according the different grid functions they contain.
- There are two :py:class`~.OneGridFunction` classes, one for HDF5 files and one
  for ASCII files. They describe one single grid function and they contains the
  files associated to that grid function. Both the classes are derived from the
  same abstract base class :py:class`~.OneGridFunctionBase`, which implements
  the shared methods.

These are hierarchical classes, one containing the others, so one typically ends
up with a series of brackets to access the actual data. For example, if ``sim``
is a :py:class:`~.SimDir`, ``sim.gf.xy['rho_b'][0]`` is ``rho_b`` at iteration 0
on the equatorial plane represented as :py:class:`~.HierarchicalGridData`.

"""

import os
import re
import warnings
from abc import ABC, abstractmethod
from bz2 import open as bopen
from contextlib import contextmanager
from functools import lru_cache
from gzip import open as gopen

import h5py
import numpy as np

from kuibit import grid_data, simdir
from kuibit.attr_dict import pythonize_name_dict
from kuibit.cactus_ascii_utils import scan_header, total_filesize


class BaseOneGridFunction(ABC):
    """Abstract class that implements capabilities to handle grid functions.

    This class is the parent class of :py:class:`~.OneGridFunctionASCII` and
    :py:class:`~.OneGridFunctionH5`. :py:class:`~.BaseOneGridFunction`
    implements most methods, except the readers.

    The derived classes have to specify:

    - How to read a file, populating the ``self.alldata`` dictionary
      (method ``_parse_file``).
    - How to populate the last level of ``self.alldata`` by returning
      a :py:class:`~.UniformGridData` for a given iteration and component
      (method ``_read_componenent_as_uniform_grid``)
    - How to associate an iteration with a time (method
      ``time_at_iteration``).

    The simplest way to access data at a given iteration as
    :py:class:`~.HierarchicalGridData` is using the ``[]`` notation.

    :ivar allfiles: Paths of files associated to the variable.
    :type allfiles: list of str
    :ivar alldata: Dictionary that organizes files and iterations available.
    :type alldata: nested dictionary
    :ivar restarts_data: How iterations are distributed across files.
    :type restarts_data: tuple of str
    :ivar var_name: Variable name.
    :type var_name: str

    """

    def __init__(self, allfiles, var_name):
        """Constructor.

        :param allfiles: Paths of files associated to the variable.
        :type allfiles: list of str
        :param var_name: Variable name.
        :type var_name: str

        """

        self.allfiles = list(allfiles)

        # self.alldata is a nested dictionary
        # 1. At the first level, we have the file
        # 2. self.alldata[filename] is a dictionary with keys the various
        #    available iterations in filename
        # 3. self.alldata[filename][iteration] is another dictionary with keys the
        #    various refinement levels available in filename at the iteration
        #    and as values another dictionary
        # 4. This last dictionary has as keys the available components and as
        #    values None if the data has not been read yet, or UniformGridData
        #    for that component if the data has been read
        self.alldata = {}

        # We use this to extract only the information related to the specific
        # variable
        self.var_name = var_name

        for path in self.allfiles:
            self._parse_file(path)

        # Here we are going to save the restart information
        self.restarts_data = None

    # The derived classes have to specify:
    # 1. How to read a file, populating the self.alldata dictionary up to the
    #    last level (excluded or included) (_parse_file). This should also
    # 2. How to populate the last level of self.alldata by returning
    #    a UniformGridData for a given iteration and component
    #    (_read_componenent_as_uniform_grid)
    # 3. How to associate an iteration with a time (time_at_iteration)

    @abstractmethod
    def _parse_file(self, path):
        """Read file at path and populate ``self.alldata``."""
        raise NotImplementedError

    @abstractmethod
    def _read_component_as_uniform_grid_data(
        self, path, iteration, ref_level, component
    ):
        """Read specific component."""
        raise NotImplementedError

    @abstractmethod
    def time_at_iteration(self, iteration):
        """Return the time at a given iteration.

        :param iteration: Iteration.
        :type iteration: int
        :returns: Time.
        :rtype: float
        """
        raise NotImplementedError

    @lru_cache(128)
    def _iterations_in_file(self, path):
        """Return the (sorted) available iterations in file path.

        Use this if you need to ensure that you are looping over iterations in
        order!

        :param path: File to inspect.
        :type path: str

        :returns: Sorted list of iterations available in a given file.
        :rtype: list

        """
        return sorted(list(self.alldata[path].keys()))

    def _min_iteration_in_file(self, path):
        """Return the minimum available iterations in the given file.

        :param path: File to inspect.
        :type path: str

        :returns: First iteration available.
        :rtype: int

        """
        return self._iterations_in_file(path)[0]

    def _max_iteration_in_file(self, path):
        """Return the maximum available iterations in the given file.

        :param path: File to inspect.
        :type path: str

        :returns: Last iteration available.
        :rtype: int

        """
        return self._iterations_in_file(path)[-1]

    def _get_restarts(self):
        """Return a list of tuples of the form ``(iteration_min, iteration_max, paths)``
        with the minimum iterations and maximum iterations in each file
        associated to this variable. ``paths`` is a list of all the files with
        same ``min_iteration`` and ``max_iteration`` (as when we have 3d
        xyz_file files).

        This routine is used to identify which files to use for any given
        iteration.

        :returns: List of tuples with first iteration, last iteration, and
                  path for every file in ``self.allfiles``.
        :rtype: list of tuple (int, int, list of str)

        """
        restarts = [
            (
                self._min_iteration_in_file(path),
                self._max_iteration_in_file(path),
                [path],
            )
            for path in self.allfiles
        ]
        # We sort by first the minimum iteration. If the minimum iteration is
        # the same, we sort by the -maximum iteration, which ensures that where
        # we have more iterations is placed before. Consider the example of
        # list to be sorted [(1, 2), (0, 3), (1, 5)], this would be sorted in
        # ascending order with respect as if it was [(1, -2), (0, -3), (1, -5)],
        # so [(0, 3), (1, 5), (1, 2)]
        restarts.sort(key=lambda x: (x[0], -x[1]))

        # Next, we check that there is no overlap. If there's overlap, we
        # ignore some folders, unless min_iteration and max_iteration are
        # exactly the same, in which case we combine the two
        first, *others = restarts
        # We assemble a return list, starting with the first element. Then,
        # we loop over the other elements and we keep expanding the return
        # list only if we find that the maximum iteration is larger than
        # the previous maximum iteration. Since we sorted as described above,
        # this will ignore repeated runs in which the iterations are a subset
        # of the previous one.
        ret = [first]
        for min_iteration, max_iteration, path in others:
            # If we have exactly the same iterations, we are looking at ones
            # of those 3D files, so we collect them
            if (min_iteration, max_iteration) == (ret[-1][0], ret[-1][1]):
                ret[-1][2].append(path)
                continue

            max_iteration_in_ret = ret[-1][1]
            if max_iteration > max_iteration_in_ret:
                ret.append((min_iteration, max_iteration, path))
            else:
                warnings.warn(f"Unused (redundant) file: {path}")

        return ret

    @property
    def restarts(self):
        """Return a list of tuples of the form ``(iteration_min, iteration_max, paths)``
        with the minimum iterations and maximum iterations in each file
        associated to this variable. ``paths`` is a list of all the files with
        same ``min_iteration`` and ``max_iteration`` (as when we have 3d
        xyz_file files).

        :returns: List of tuples with first iteration, last iteration, and
                  path for every file in ``self.allfiles``.
        :rtype: list of tuple (int, int, list of str)
        """
        if self.restarts_data is None:
            self.restarts_data = self._get_restarts()
        return self.restarts_data

    @property
    def min_iteration(self):
        """Return the minimum available iteration in the all the files.

        :returns: First iteration available.
        :rtype: int

        """
        # restarts is a ordered list of tuples with three elements:
        # (iteration_min, iteration_max, path)

        # The minimum iteration is in restarts[0][0]
        return self.restarts[0][0]

    @property
    def max_iteration(self):
        """Return the maximum available iteration in the all the files.

        :returns: Latest iteration available.
        :rtype: int

        """
        # restarts is a ordered list of tuples with three elements:
        # (iteration_min, iteration_max, path)

        # The maximum iteration is in restarts[-1][1]
        return self.restarts[-1][1]

    @property
    @lru_cache(128)
    def available_iterations(self):
        """Return the available iterations.

        :returns: List with all the available iterations.
        :rtype: list

        """
        iterations_in_files = set()
        for path in self.allfiles:
            iterations_in_files.update(self._iterations_in_file(path))

        # Next we merge everything to make a set and we sort it
        return sorted(list(iterations_in_files))

    @property
    @lru_cache(128)
    def available_times(self):
        """Return the available times.

        :returns: List with all the available times.
        :rtype: list

        """
        return [
            self.time_at_iteration(iteration)
            for iteration in self.available_iterations
        ]

    times = available_times
    iterations = available_iterations

    def __iter__(self):
        for iteration in self.available_iterations:
            yield self[iteration]

    def iteration_at_time(self, time):
        """Return the iteration that corresponds to the given time.

        :param time: Time.
        :type time: float

        :returns: Iteration corresponding to the given time.
        :rtype: int

        """

        # TODO (FEATURE): Add tolerance in time.
        #
        # Often different outputs are not synced, so we need to allow for some
        # tolerance.

        if time not in self.available_times:
            raise ValueError(f"Time {time} not available")

        index = np.searchsorted(self.available_times, time)
        return self.available_iterations[index]

    def _ref_levels_in_file(self, path, iteration):
        """Return the available refinement levels in the given path at the
        specified iteration.

        :param path: Path of the file.
        :type path: str
        :param iteration: Iteration.
        :type iteration: int

        :returns: Available refinement levels in ``path`` at ``iteration``.
        :rtype: list
        """
        # In case we don't have the iteration, we want to return the empty
        # list, so we use the .get dictionary method. This return the
        # value, if the key is available, otherwise it returns the second
        # argument. Here, we fix as second argument the empty dictionary, so
        # when we take .keys() we get the empty list
        return list(self.alldata[path].get(iteration, {}).keys())

    def _components_in_file(self, path, iteration, ref_level):
        """Return the available components in the given file at the specified iteration
        and refinement level.

        :param path: Path of the file.
        :type path: str
        :param iteration: Iteration.
        :type iteration: int
        :param ref_level: Refinement level.
        :type ref_level: int

        :returns: Available components in ``path`` at ``iteration`` and ``ref_level``.
        :rtype: list

        """
        # Same comment as _ref_levels_in_file, but with an
        # additional level
        return list(
            self.alldata[path].get(iteration, {}).get(ref_level, {}).keys()
        )

    def _files_with_iteration(self, iteration):
        """Return all the files that contain the given iteration.

        :param path: Path of the file.
        :type path: str
        :param iteration: Iteration.
        :type iteration: int

        :returns: Files that contain the given iteration.
        :rtype: list

        """
        # Using self.get_restarts(), find the file that the given iteration
        # between iteration_min and iteration_max

        if (iteration < self.min_iteration) or (
            iteration > self.max_iteration
        ):
            raise ValueError(f"Iteration {iteration} not in available range")

        # restarts is a ordered list of tuples with three elements:
        # (iteration_min, iteration_max, path)
        max_iterations_in_files = np.array([i for _, i, _ in self.restarts])
        # This returns the index of the first element in max_iterations_in_files
        # that is greater than the given iteration.
        # E.g. max_iterations_in_files = [1, 5, 10] and iteration = 6, since
        # max_iterations_in_files are the maximum iterations in the various files,
        # the function has to return 2, which is the index of the first element
        # in max_iterations_in_files that is larger than 6. This is what
        # np.searchsorted does.
        index = np.searchsorted(max_iterations_in_files, iteration)

        # The second element is the list of path
        return self.restarts[index][2]

    def total_filesize(self, unit="MB"):
        """Return the total size of the files with this variable.
        Available units B, KB, MB and GB (in power of 1024 bytes).

        :param unit: Unit to use (in powers of 1024 bytes).
        :type unit: str among: ``B``, ``KB``, ``MB``, ``GB``.
        :returns: Total size of the files associated to this variable.
        :rtype: float

        """
        return total_filesize(self.allfiles, unit=unit)

    def _read_iteration_as_HierarchicalGridData(self, iteration):
        """Return the data at the given iteration as a :py:class:`~.HierarchicalGridData`.

        :param iteration: Iteration.
        :type iteration: int

        :returns: Variable at the given iteration as a
                  :py:class:`~.HierarchicalGridData`.
        :rtype: :py:class:`~.HierarchicalGridData`

        """

        uniform_grid_data_components = []

        for path in self.allfiles:
            for ref_level in self._ref_levels_in_file(path, iteration):
                for comp in self._components_in_file(
                    path, iteration, ref_level
                ):
                    uniform_grid_data_components.append(
                        self._read_component_as_uniform_grid_data(
                            path, iteration, ref_level, comp
                        )
                    )

        return (
            grid_data.HierarchicalGridData(uniform_grid_data_components)
            if uniform_grid_data_components
            else None
        )

    def get_iteration(self, iteration, default=None):
        """Return the data at the given iteration as a :py:class:`~.HierarchicalGridData`.
        If the iteration is not available, return ``default``.

        :param iteration: Iteration.
        :type iteration: int
        :param default: What to return if iteration is not available.
        :type default: anything

        :returns: Variable at the given iteration as a
                  :py:class:`~.HierarchicalGridData`.
        :rtype: :py:class:`~.HierarchicalGridData`

        """

        if iteration not in self.available_iterations:
            return default
        return self[iteration]

    def get_time(self, time, default=None):
        """Return the data at the given time as a :py:class:`~.HierarchicalGridData`.
        If the time is not available, return ``default``.

        :param iteration: Iteration.
        :type iteration: int
        :param default: What to return if time is not available.
        :type default: anything

        :returns: Variable at the given time as a
                  :py:class:`~.HierarchicalGridData`.
        :rtype: :py:class:`~.HierarchicalGridData`

        """
        if time not in self.available_times:
            return default
        return self[self.iteration_at_time(time)]

    def __getitem__(self, iteration):
        if iteration not in self.available_iterations:
            raise KeyError(f"Iteration {iteration} not present")

        return self._read_iteration_as_HierarchicalGridData(iteration)

    def read_on_grid(self, iteration, grid, resample=False):
        """Read an iteration and resample the output on the specified grid.

        Warning: this can be computationally expensive!

        :param iteration: requested iteration
        :type iteration: time
        :param grid:
        :type grid: UniformGrid
        :param resample: Whether to use multilinear interpolation
        :type resample: bool
        """
        return self[iteration].to_UniformGridData_from_grid(
            grid, resample=resample
        )

    # def read_evolution_on_grid(
    #     self,
    #     grid,
    #     read_every=None,
    #     min_iteration=None,
    #     max_iteration=None,
    #     **kwargs,
    # ):
    #     """Read multiple iterations at once on the specified grid and return the result
    #     as RegularGridData in which the first index is the time and the other indices
    #     are the spatial indices of grid.

    #     """
    #     # TODO: IMPLEMENT THIS

    #     iterations = self.available_iterations
    #     if min_iteration is not None:
    #         iterations = iterations[iterations >= min_iteration]
    #     if max_iteration is not None:
    #         iterations = iterations[iterations <= max_iteration]
    #     if read_every is None:
    #         read_every = np.diff(iterations).max()

    #     iterations = [
    #         i for i in iterations if ((i - iterations[0]) % read_every == 0)
    #     ]
    #     times = [self.time_at_iteration(i) for i in iterations]

    #     dt = np.diff(times).min()
    #     if dt <= 0:
    #         raise RuntimeError("Non-positive timesteps detected.")
    #     #
    #     if abs(np.diff(times).max() - dt).max() > dt * 1e-5:
    #         raise RuntimeError("Timestep not constant enough")

    #     data = np.asarray(
    #         [self.read_on_grid(i, grid).data for i in iterations]
    #     )

    #     new_x0 = [times[0]] + list(grid.x0)
    #     new_dx = [dt] + list(grid.dx)

    #     return grid_data.UniformGridData.from_grid_structure(
    #         data, x0=new_x0, dx=new_dx
    #     )


class OneGridFunctionASCII(BaseOneGridFunction):
    """Read grid data produced by CarpetASCII.

    This class is derived from :py:class:`~.BaseOneGridFunction` and implements
    the reading facilities.

    :py:class:`~.OneGridFunctionASCII` can read 1D, 2D, and 3D ASCII files, even
    when they are compressed with bzip2 or gzip.

    ASCII files do not contain information about the ghost zones, but this can be
    set "by hand".

    """

    # TODO (REFACTORING): Avoid reading files twice.
    #
    # This class has to read the all files. When there are multiple variables in
    # one file, it would be better to avoid reading again the various files (if
    # we have already read them). Maybe we can add another level in the class
    # hierarchy that contains all the information for a given file, and produces
    # transparently OneGridFunctionASCII objects upon request.

    # What function to use to open the file?
    # What mode?
    _decompressor = {
        None: (open, "r"),
        "gz": (gopen, "rt"),
        "bz2": (bopen, "rt"),
    }

    def __init__(self, allfiles, var_name, num_ghost=None):
        """Constructor.

        :param allfiles: Paths of files associated to the variable.
        :type allfiles: list of str
        :param var_name: Variable name.
        :type var_name: str
        :param num_ghost: Number of ghost zones in each direction.
        :type num_ghost: 1d NumPy array

        """

        self._iterations_to_times = {}
        self.num_ghost = num_ghost

        super().__init__(allfiles, var_name)

    def _parse_file(self, path):
        """Read the content of the given file.

        At the moment, we read the entire file line by line, which is very
        inefficient.

        :param path: Path of the file to read.
        :type path: str

        """

        # First we parse the header to find the column description, then we read
        # the ENTIRE file. This is very inefficient, but it is not too hard to
        # implement.

        # This regex is meant to understand if we have one variable per file or
        # one group per file, and to understand if we have compression. To see a
        # detailed explanation, see AllGridFunctions. The only difference here
        # is that we don't care about the extension, so we have an addition (*)?
        rx_filename = re.compile(
            r"^(([a-zA-Z0-9_]+)-)?([a-zA-Z0-9\[\]_]+).([xyz]+)?.asc(\.(gz|bz2))?$"
        )

        filename = os.path.split(path)[1]
        matched = rx_filename.match(filename)

        if matched is None:
            raise RuntimeError(f"Found file with unusual name: {path}")

        is_one_file_per_group = matched.group(1) is not None

        compression_method = matched.group(6)
        opener, opener_mode = OneGridFunctionASCII._decompressor[
            compression_method
        ]

        # These files always have the column format line, and have the data
        # format line only if they are "one file per group"
        _, column_description = scan_header(
            path,
            one_file_per_group=is_one_file_per_group,
            extended_format=True,
            opener=opener,
            opener_mode=opener_mode,
        )
        # We have two possibilities, one is that the file only contains one
        # variable, column_description will be the column number. If the
        # file contains many variables, column_description is a dictionary
        # that maps variables to their column.
        if isinstance(column_description, dict):
            # The variable we work with is column_description so we overwrite it
            # to be the number of column with the data we are interested in
            column_description = column_description[self.var_name]

        # Now we read the entire file, line by line. This is the most
        # inefficient way possible. But at least, it is reasonably
        # straightforward to implement. If you are reading this comment and you
        # want to improve this, feel free to do it.

        def current_data_to_UniformGridData(
            current_x,
            current_y,
            current_z,
            current_data,
            current_time,
            current_iteration,
            current_component,
            current_ref_level,
        ):
            # First, we compute x0 and x1
            x0_3d = np.asarray(
                [
                    np.amin(current_x),
                    np.amin(current_y),
                    np.amin(current_z),
                ]
            )
            x1_3d = np.array(
                [
                    np.amax(current_x),
                    np.amax(current_y),
                    np.amax(current_z),
                ]
            )

            # Now we find the interesting dimensions
            dimensions_in_data = x0_3d != x1_3d

            # With unique we find the real data
            shape_3d = [
                len(np.unique(current_x)),
                len(np.unique(current_y)),
                len(np.unique(current_z)),
            ]

            shape = np.asarray(shape_3d)[dimensions_in_data]
            x0 = np.asarray(x0_3d)[dimensions_in_data]
            x1 = np.asarray(x1_3d)[dimensions_in_data]

            var_data = np.array(current_data).reshape(tuple(shape[::-1]))

            grid = grid_data.UniformGrid(
                shape,
                x0=x0,
                x1=x1,
                num_ghost=self.num_ghost,
                component=current_component,
                ref_level=current_ref_level,
                time=current_time,
                iteration=current_iteration,
            )

            return grid_data.UniformGridData(grid, np.transpose(var_data))

        # We are going to assume that the iteration column is the first
        with opener(path, opener_mode) as fil:
            # We use these variables as local variables. We are going to aggregate
            # the data read here, and reset it when we find a blank line
            current_iteration = None
            current_ref_level = None
            current_component = None
            current_x = []
            current_y = []
            current_z = []
            current_data = []
            # We scan the entire file line by line
            for line in fil:
                # Skip header
                if not line[0].isdigit():
                    # We don't care about lines that don't start with a number
                    continue

                # Here are can assume that this is a line with data
                line_data = line.split()
                line_data = list(map(float, line_data))
                if current_iteration is None:
                    current_iteration = line_data[0]

                if current_ref_level is None:
                    current_ref_level = line_data[2]

                if current_component is None:
                    current_component = line_data[3]

                # If iteration, component, or refinement level changes, we
                # write the data, else we continue reading
                if (
                    line_data[0] != current_iteration
                    or line_data[2] != current_ref_level
                    or line_data[3] != current_component
                ):
                    alldata_file = self.alldata.setdefault(path, {})
                    alldata_iteration = alldata_file.setdefault(
                        int(current_iteration), {}
                    )
                    alldata_ref_level = alldata_iteration.setdefault(
                        int(current_ref_level), {}
                    )

                    current_time = line_data[8]

                    uniform_grid_data = current_data_to_UniformGridData(
                        current_x,
                        current_y,
                        current_z,
                        current_data,
                        current_time,
                        current_iteration,
                        current_component,
                        current_ref_level,
                    )

                    alldata_ref_level.setdefault(
                        int(current_component), uniform_grid_data
                    )

                    # Write iterations_to_time
                    if current_iteration not in self._iterations_to_times:
                        self._iterations_to_times[
                            current_iteration
                        ] = current_time

                    # Reset everything
                    current_iteration = line_data[0]
                    current_ref_level = line_data[2]
                    current_component = line_data[3]
                    current_x = []
                    current_y = []
                    current_z = []
                    current_data = []

                # We still have to read the data on this line even if we
                # "are done with a group"
                current_x.append(line_data[9])
                current_y.append(line_data[10])
                current_z.append(line_data[11])

                current_data.append(line_data[column_description])

            # Here we take care of the last piece of data
            if len(current_data) > 0:
                if current_iteration not in self._iterations_to_times:
                    current_time = line_data[8]
                    self._iterations_to_times[current_iteration] = current_time

                alldata_file = self.alldata.setdefault(path, {})
                alldata_iteration = alldata_file.setdefault(
                    int(current_iteration), {}
                )
                alldata_ref_level = alldata_iteration.setdefault(
                    int(current_ref_level), {}
                )

                uniform_grid_data = current_data_to_UniformGridData(
                    current_x,
                    current_y,
                    current_z,
                    current_data,
                    current_time,
                    current_iteration,
                    current_component,
                    current_ref_level,
                )

                alldata_ref_level.setdefault(
                    int(current_component), uniform_grid_data
                )

    def _read_component_as_uniform_grid_data(
        self, path, iteration, ref_level, component
    ):
        """Return the component at the given iteration, refinement level, and component.

        :param path: Path of the file.
        :type path: str
        :param iteration: Iteration.
        :type iteration: int
        :param ref_level: Refinement level.
        :type ref_level: int
        :param component: Component.
        :type component: int

        :returns: Component as a :py:class:`~.UniformGridData`.
        :rtype: :py:class:`~.UniformGridData`

        """

        # We have already read the files, so we just return it
        return self.alldata[path][iteration][ref_level][component]

    def time_at_iteration(self, iteration):
        """Return the time at a given iteration.

        :param iteration: Iteration.
        :type iteration: int

        :returns: Time at given iteration.
        :rtype: float

        """

        if iteration not in self.available_iterations:
            raise ValueError("Iteration {iteration} not available")

        return self._iterations_to_times[iteration]


class OneGridFunctionH5(BaseOneGridFunction):
    """Read grid data produced by CarpetHDF5 files.

    This class is derived from :py:class:`~.BaseOneGridFunction` and implements
    the reading facilities.

    """

    # This class implements the details on how to read the data, most of the
    # functionalities of the class are in OneGridFunctionBase.

    # Let's unpack the regex, we have 7 capturing groups
    #
    # 1. [^:] matches any number of characters (at least one), with the
    #    exception of ':'
    # 2. \S+ matches any number of non-whitespace characters (at least
    #    one).
    #
    # 1. and 2. are thorn and variable name, then we have
    # 3, 4. \d+ matches a number (iteration and time level)
    # 5. " m=0" matches this exactly, if it is present
    # 6. \d+ matches a number (refinement level) if present
    #    (grid arrays don't have this)
    # Then we have two nested capturing groups
    # ( c=(\d+))? checked whether c=NUMBER is matched,
    # and inside we have that the component number is matched
    #
    # All the [ ] are meaningful blank spaces. Note, we don't have $^ because
    # there is more that we do not want to match
    _pattern_group_name = r"""
    ([^:]+)             # Thorn name
    ::
    (\S+)               # Variable name
    [ ]
    it=(\d+)            # Iteration
    [ ]
    tl=(\d+)            # Timelevel (almost always 0)
    ([ ]m=0)?           # Map
    ([ ]rl=(\d+))?      # Refinement level
    ([ ]c=(\d+))?       # Component
    """

    def __init__(self, allfiles, var_name: str, dimension):
        """Constructor.

        :param allfiles: Paths of files associated to the variable.
        :type allfiles: list of str
        :param var_name: Variable name.
        :type var_name: str

        """

        # We need these variables to properly find what dataset to look at in
        # the HDF5 file.
        self.thorn_name = None
        self.map = None

        self.dimension = dimension

        self.rx_group_name = re.compile(self._pattern_group_name, re.VERBOSE)

        super().__init__(allfiles, var_name)

        # super() will fill the other variables that we need for dataset_format
        if self.map is None:
            self.map = ""

        # TODO (FEATURE): Make separator customizable
        #
        # Technically the separator :: is customizable, so we should be more
        # flexible.

        self.dataset_format = (
            f"{self.thorn_name}::{self.var_name} it=%d tl=0{self.map}%s%s"
        )

        file_is_3D = len(self.dimension) == 3

        # HDF5 files can contain ghostzones or not. Here, we can that all the
        # files have the same behavior (they all contain, or they all don't)
        #
        # self._are_ghostzones_in_file(path) returns True or False, so this
        # is a set with True, False or a mix
        ghost_in_files = {
            self._are_ghostzones_in_file(path, file_is_3D) for path in self.allfiles
        }

        # Here we check that we only have True or False
        if len(ghost_in_files) != 1:
            raise ValueError(
                "Inconsistent IOHDF5::output_ghost_points across files"
            )

        # We know that ghost_in_files has only one element (either True or
        # False), so we pick that (with tuple unpacking)
        (self.are_ghostzones_in_files,) = ghost_in_files

    def _parse_file(self, path):
        """Read the content of the given file (without reading the data).

        :param path: Path of the file.
        :type path: str

        """
        # This will give us an overview of what is available in the provided
        # file. We keep a collection of all these in the variable self.alldata
        with h5py.File(path, "r") as f:
            for group in f.keys():
                matched = self.rx_group_name.match(group)
                # If this is not an interesting group, just skip it
                if not matched:
                    continue

                (
                    thorn_name,
                    var_name,
                    iteration,
                    time_level,
                    map_,
                    _,
                    ref_level,
                    _,
                    component,
                ) = matched.groups()

                if var_name != self.var_name:
                    continue

                time_level = int(time_level)

                # We only care about the current timelevel
                if time_level != 0:
                    continue

                if self.thorn_name is None:
                    self.thorn_name = thorn_name

                if self.map is None:
                    self.map = map_

                component = -1 if matched.group(9) is None else int(component)
                # This is important to support grid arrays, which do not have a
                # refinement level
                ref_level = -1 if matched.group(7) is None else int(ref_level)

                # Here is where we prepare are nested alldata dictionary
                alldata_file = self.alldata.setdefault(path, {})
                alldata_iteration = alldata_file.setdefault(int(iteration), {})
                alldata_ref_level = alldata_iteration.setdefault(ref_level, {})

                # We set the actual data to None, and we will read it in
                # _read_component_as_uniform_grid_data upon request
                alldata_ref_level.setdefault(int(component), None)

    def _grid_from_dataset(self, dataset, iteration, ref_level, component):
        """Return a :py:class:`~.UniformGrid` from a given HDF5 dataset.

        :param dataset: Dataset to model the grid after.
        :type dataset: H5py.dataset
        :param iteration: Iteration.
        :type iteration: int
        :param ref_level: Refinement level.
        :type ref_level: int
        :param component: Component.
        :type component: int

        :returns: Grid corresponding to the dataset.
        :rtype: :py:class:`~.UniformGrid`

        """

        # NOTE: Why are we taking the reverse?
        shape = np.array(dataset.shape[::-1])

        x0 = dataset.attrs["origin"]
        dx = dataset.attrs["delta"]
        time = dataset.attrs["time"]
        # With the .get we ensure that if "cctk_nghostzones" cannot be read, we
        # have returned None, which we can test later
        num_ghost = dataset.attrs.get("cctk_nghostzones", None)
        # If we do not have the ghostzones in the file, then it is as if we
        # have ghostzones of size zero.
        if not self.are_ghostzones_in_files or num_ghost is None:
            num_ghost = np.zeros_like(shape, dtype=int)

        return grid_data.UniformGrid(
            shape,
            x0=x0,
            dx=dx,
            ref_level=ref_level,
            num_ghost=num_ghost,
            time=time,
            iteration=iteration,
            component=component,
        )

    # What is a context manager?
    #
    # Context managers are useful ways to handle resources in Python. With a
    # context manager, we do not have to worry about releasing resources. Here,
    # we wrap reading the h5 file with another context manager so that we can
    # easily get the dataset.
    @contextmanager
    def _get_dataset(self, path, iteration, ref_level, component):
        """Context manager to read an HDF5 file.

        :param path: Path of the file.
        :type path: str
        :param iteration: Iteration.
        :type iteration: int
        :param ref_level: Refinement level.
        :type ref_level: int
        :param component: Component.
        :type component: int

        """
        ref_level_str = f" rl={ref_level}" if (ref_level >= 0) else ""
        component_str = f" c={component}" if (component >= 0) else ""
        with h5py.File(path, "r") as f:
            try:
                yield f[
                    self.dataset_format
                    % (iteration, ref_level_str, component_str)
                ]
            finally:
                # All the hard work is done by the other 'with' statement.
                # We don't need to do anything here.
                pass

    def _read_component_as_uniform_grid_data(
        self, path, iteration, ref_level, component
    ):
        """Return the component at the given iteration, refinement level, and component.

        :param path: Path of the file.
        :type path: str
        :param iteration: Iteration.
        :type iteration: int
        :param ref_level: Refinement level.
        :type ref_level: int
        :param component: Component.
        :type component: int

        :returns: Component as a :py:class:`~.UniformGridData`.
        :rtype: :py:class:`~.UniformGridData`

        """

        if self.alldata[path][iteration][ref_level][component] is None:
            with self._get_dataset(
                path, iteration, ref_level, component
            ) as dataset:
                grid = self._grid_from_dataset(
                    dataset, iteration, ref_level, component
                )
                data = np.transpose(dataset[()])

                self.alldata[path][iteration][ref_level][
                    component
                ] = grid_data.UniformGridData(grid, data)

        return self.alldata[path][iteration][ref_level][component]

    @staticmethod
    def _are_ghostzones_in_file(path: str, is_3D_file: bool) -> bool:
        """Return whether the ghostzones were output or not.

        :param path: File to inspect.
        :type path: str
        :param is_3D_file: The file contains 3D data
        :type is_3D_file: bool

        :returns: Whether ``path`` contains ghost zones.
        :rtype: bool

        """
        # This is a tricky and important function to stitch together all the
        # different components. Carpet has an option (technically two) to output
        # the ghostzones in the files. These are: output_ghost_points and
        # out3D_ghosts (which is deprecated). For 3D files, when they are both
        # set to yes, the ghostzones are output in the h5 files. When one of the
        # two is set to no, the ghostzones are not output. For 2D files, when
        # it depends only on the value of the first.

        # The default value of these parameters is yes
        with h5py.File(path, "r") as file_:
            parameters = file_["Parameters and Global Attributes"]
            all_pars = parameters["All Parameters"][()].decode().split("\n")
            # We make sure that everything is lowercase, we are case insensitive
            iohdf5_pars = [
                param.lower()
                for param in all_pars
                if param.lower().startswith("carpetiohdf5")
                or param.lower().startswith("iohdf5")
            ]

            def is_param_true(name: str) -> bool:
                param = [p for p in iohdf5_pars if name.lower() in p]
                # When the parameters we are interested in are not set, they are
                # set to yes by default
                if len(param) == 0:
                    return True

                # The parameter is set
                return (
                    ("true" in param[0])
                    or ("yes" in param[0])
                    or ("1" in param[0])
                )

            params_to_check = {"output_ghost_points"}

            if is_3D_file:
                params_to_check.add("out3D_ghosts")

            # Ghostzones are in the file if all the relevant parameters are set
            # to true
            return all(is_param_true(param) for param in params_to_check)

    def clear_cache(self):
        """Remove all the cached entries.

        Every time a component is read, :py:class:`~.OneGridFunctionsH5` caches
        its value (reading can be expensive). In certain cases, this can lead to
        an explosion in the size of this object. For example, when reading
        several iterations to make a movie. This method removes all the cached
        entries, keeping the size of the object under control.
        """
        for filename, file_reader in self.alldata.items():
            for iteration, iteration_reader in file_reader.items():
                for ref_level, ref_level_reader in iteration_reader.items():
                    for component in ref_level_reader.keys():
                        self.alldata[filename][iteration][ref_level][
                            component
                        ] = None

    def time_at_iteration(self, iteration):
        """Return the time corresponding to the provided iteration.

        :param iteration: Iteration.
        :type iteration: int

        :returns: Time corresponding to ``iteration``.
        :rtype: float

        """
        # If there are multiple files, we take the first.
        # A case in which there are multiple files is with 3D data
        path = self._files_with_iteration(iteration)[0]

        ref_levels = self._ref_levels_in_file(path, iteration)
        components = self._components_in_file(path, iteration, ref_levels[-1])
        with self._get_dataset(
            path, iteration, ref_levels[-1], components[-1]
        ) as dataset:
            return dataset.attrs["time"]


class AllGridFunctions:
    """Helper class to read various types of grid data in a list of files and
    properly order them. The core of this object is the ``_vars`` dictionary
    which contains the location of all the files for a specific variable and
    reduction.

    :py:class:`~.AllGridFunction` is a dictionary-like object with keys the
    various variables and values :py:class:`~.BaseOneGridFunction` (or derived).

    You can access data with the bracket operator or as attributes of the
    ``fields`` attribute.

    Not intended for direct initialization.

    :ivar dimension: Dimension associated to this object (e.g. (0, 1) would be the
                     xy plane).
    :type dimension: tuple
    :ivar num_ghost: Number of ghost zones in each dimension.
    :type num_ghost: 1d NumPy array.

    """

    # Different "cuts" have different extensions in their filenames, here we
    # save all the possible ones. In the instance of the class, we fix the
    # specific pattern corresponding to the dimension (which are the keys on
    # the following dictionary). In general, the file name will be:
    # variable-name.ext.h5, eg rho.xy.h5.
    filename_extensions = {
        (0,): ".x",
        (1,): ".y",
        (2,): ".z",
        (0, 1): "(.0)?.xy",
        (0, 2): "(.0)?.xz",
        (1, 2): "(.0)?.yz",
        (0, 1, 2): r"(.xyz)?(.file_[\d]+)?(.xyz)?",
    }

    _dim_names = {
        (0,): "x",
        (1,): "y",
        (2,): "z",
        (0, 1): "xy",
        (0, 2): "xz",
        (1, 2): "yz",
        (0, 1, 2): "xyz",
    }

    def __init__(self, allfiles, dimension, num_ghost=None):
        """Constructor.

        :param allfiles: List of all the files.
        :type allfiles: list
        :param dimension: Dimension associated to this object.
        :type dimension: tuple
        :param num_ghost: Number of ghost zones in the data for each dimension.
                          This is used only for ASCII data.
        :type num_ghost: list or tuple of the same length as the number of dimension

        """

        # Here we save what kind of file we are looking at
        # We assume that dimension is already sanitized (that is, is in tuple
        # form and not in string form)
        self.dimension = dimension

        # If we are using ASCII files, we have to know how many ghost zones are
        # in the data. At the moment we ask the user to provide the data, but
        # in the future we will parse the paramter file and find this value.
        #
        # We don't use this value for HDF5 data, as it is more reliable to just
        # read it from the files.
        # Here we are using a setter for num_ghost, see below
        self.num_ghost = num_ghost

        # This is a simple regex:
        # 1. ^ and $ mean that we have to match the entire string
        # 2. ([a-zA-Z0-9_]+) means that we match any combination of letters
        #    and numebrs. This is the thorn name when we output one group
        #    per file. We wrap this into another capturing group:
        #    (([a-zA-Z0-9_]+)-). Here we also try to match the literal '-'.
        #    This separates the thron name from the group name. If we match
        #    this larger capturing group, it means that the file was output
        #    with the option "one_group_per_file".
        # 3. ([a-zA-Z0-9\[\]_]+) means that we match any character any number
        #    of times, this is a capturing group, and is the variable name,
        #    or the group name if we output one group per file.
        # 4. We have the extension, which identifies the dimension and is
        #    saved in the class instance
        # 5. Finally, we have the filename extension which can be either h5
        #    or txt
        #
        # Example of filenames are:
        # admbase-metric.xyz.file_158.h5 (one group per file)
        # alp.xy.h5 (one variable per file)
        # filename_pattern = r"^([a-zA-Z0-9_]+)(-)?([a-zA-Z0-9\[\]_]+)%s.%s$"
        filename_pattern = r"^(([a-zA-Z0-9_]+)-)?([a-zA-Z0-9\[\]_]+)%s.%s$"
        h5_pattern = filename_pattern % (
            self.filename_extensions[self.dimension],
            "h5",
        )
        ascii_pattern = filename_pattern % (
            self.filename_extensions[self.dimension],
            r"asc(\.(gz|bz2))?",
        )

        # Variable files is a dictionary, the keys are the variables, the
        # values the set of files associated to that variable
        self._vars_ascii_files = {}
        self._vars_h5_files = {}

        # _vars contains the actual data. It is used to cache results. _vars is
        # a dictionary with keys the variables and values OneGridFunction (H5 or
        # ASCII)
        self._vars = {}

        rx_h5 = re.compile(h5_pattern)
        rx_ascii = re.compile(ascii_pattern)

        # Here we scan all the files and find those with a name that match
        # one of our regular expressions.

        for f in allfiles:
            filename = os.path.split(f)[1]
            matched_h5 = rx_h5.match(filename)
            matched_ascii = rx_ascii.match(filename)
            # If matched_pattern is not None, this is a Carpet h5 file
            if matched_h5 is not None:
                # First, we understand if the file was output with
                # "one_group_per_file". In this case, the file contains
                # multiple variables. Files output in this way have names:
                # thorname-groupname.dim.h5 (possibly with also file_NUM).
                #
                # Group1 contains thorname- (notice the -). If we match
                # group1 it means that the file contains one group,
                # the thorn name is in group2, and the group name in group3.
                #
                # In case group1 is not matched, then the variable name is
                # in group3.
                if matched_h5.group(1) is None:
                    variable_name = matched_h5.group(3)
                    var_list = self._vars_h5_files.setdefault(
                        variable_name, set()
                    )
                    var_list.add(f)
                else:
                    # We have to open the file to understand which variables
                    # are available
                    #
                    # We use the pattern name in OneGridFunctionH5
                    rx_group_name = re.compile(
                        OneGridFunctionH5._pattern_group_name, re.VERBOSE
                    )
                    with h5py.File(f, "r") as h5f:
                        # Here group is in the sense of HDF5 group
                        for group in h5f.keys():
                            group_matched = rx_group_name.match(group)
                            # If this is not an interesting group, just skip it
                            if not group_matched:
                                continue
                            variable_name = group_matched.group(2)
                            var_list = self._vars_h5_files.setdefault(
                                variable_name, set()
                            )
                            var_list.add(f)
            elif matched_ascii is not None:
                # As in the case of H5 files, we first need to understand if
                # the output is with "one_group_per_file". If yes, we have to
                # open all the files and read the headers to find what variables
                # are available.
                if matched_ascii.group(1) is None:
                    variable_name = matched_ascii.group(3)
                    var_list = self._vars_ascii_files.setdefault(
                        variable_name, set()
                    )
                    var_list.add(f)
                else:
                    # In this case we need to open the file and scan the
                    # header, for this we use the scan_header function in
                    # cactus_scalars.
                    #
                    # Here we have to pay attention to the output of the Thorns
                    # VolumeIntegralsGRMHD and VolumeIntegralsVacuum as they
                    # produce output files with names:
                    # volume_integrals-vacuum and volume_integrals-GRMHD.
                    #
                    # These would be matched here, so we will exclude the case in
                    # which the thorn name is volume_integrals and the var name
                    # is vacuum or GRMHD.
                    #
                    # Same with the file carpet-grid.asc
                    #
                    # To deal with all of these, we wrap everything in a try
                    # except block, and keep only the variables that do not
                    # throw errors.
                    try:
                        # TODO (PERFORMANCE): Avoid reading headers twice
                        #
                        # Here we scan the headers, we should not do this work again
                        # when we deal with the single variables.

                        # The last group is where compression information is. It
                        # could be None.
                        compression_method = matched_ascii.groups()[-1]
                        (
                            opener,
                            opener_mode,
                        ) = OneGridFunctionASCII._decompressor[
                            compression_method
                        ]
                        _, column_description = scan_header(
                            f,
                            one_file_per_group=True,
                            extended_format=True,
                            opener=opener,
                            opener_mode=opener_mode,
                        )
                        for variable_name in column_description.keys():
                            var_list = self._vars_ascii_files.setdefault(
                                variable_name, set()
                            )
                            var_list.add(f)
                    except RuntimeError:
                        pass

        # What pythonize_name_dict does is to make the various variables
        # accessible as attributes, e.g. self.fields.rho
        self.fields = pythonize_name_dict(list(self.keys()), self.__getitem__)

    def __getitem__(self, key):

        var_name = str(key)

        if var_name not in self:
            raise KeyError(f"Variable {key} not present in simulation data")

        if var_name not in self._vars:
            # We prefer h5
            if var_name in self._vars_h5_files:
                self._vars[var_name] = OneGridFunctionH5(
                    self._vars_h5_files[var_name], var_name, self.dimension
                )
            elif var_name in self._vars_ascii_files:
                if self.num_ghost is None:
                    warnings.warn(
                        "You are using ASCII files, which have no information"
                        " about ghost zone information. Set the attribute num_ghost"
                        " of this object to properly account for the ghost zones. "
                    )
                self._vars[var_name] = OneGridFunctionASCII(
                    self._vars_ascii_files[var_name],
                    var_name,
                    num_ghost=self.num_ghost,
                )

        return self._vars[var_name]

    @property
    def num_ghost(self):
        """Return the number of ghost zones along each direction.

        :returns: Number of ghost zones along each direction.
        :rtype: 1d NumPy array
        """
        return self.__num_ghost

    @num_ghost.setter
    def num_ghost(self, num_ghost):
        if num_ghost is not None:
            if len(num_ghost) != len(self.dimension):
                raise ValueError(
                    "Number of ghost zones {len(num_ghost)} is inconsistent "
                    " with dimesionality (len(self.dimension))"
                )
            # We copy num_ghost, in case it was a mutable object
            self.__num_ghost = np.array(num_ghost)
        else:
            self.__num_ghost = None

    def __contains__(self, var):
        return var in self.keys()

    def get(self, key, default=None):
        """Return variable ``key``.

        :param key: Variable to read.
        :type key: str
        :param default: Value to return is variable is not available.
        :type default: anything
        """
        if key not in self:
            return default
        return self[key]

    def keys(self):
        """Return the list of all the available variables."""
        # We merge the dictionaries and return the keys.
        # This automatically takes care of making sure that they keys are unique.
        return {**self._vars_h5_files, **self._vars_ascii_files}.keys()

    def __str__(self):
        ret = "\nAvailable grid data of dimension "
        ret += f"{len(self.dimension)}D ({self._dim_names[self.dimension]}): "
        ret += f"\n{list(self.keys())}\n"
        return ret

    @property
    def allfiles(self):
        """Return a set of all the files that have variables of the
        given dimension

        :return: Collection of all the unique files with variables of
                 this dimension.
        :rtype: set
        """
        allfiles = set()
        # We collect all the files by merging the lists into a set. The
        # set will automatically remove repeated entries.
        for file_list in self._vars_h5_files.values():
            allfiles.update(file_list)
        for file_list in self._vars_ascii_files.values():
            allfiles.update(file_list)
        return allfiles

    def total_filesize(self, unit="MB"):
        """Return the total size of the files with this dimension.
        Available units B, KB, MB and GB (in power of 1024 bytes).

        :param unit: Unit to use (in powers of 1024 bytes).
        :type unit: str among: ``B``, ``KB``, ``MB``, ``GB``.
        :returns: Total size of all the files associated with this dimension.
        :rtype: float

        """
        return total_filesize(self.allfiles, unit=unit)


class GridFunctionsDir:
    """This class provides access to all grid data.

    This includes 1D-3D data in HDF5 and ASCII formats. Data of the required
    dimensionality is read from any format available (HDF5 preferred over
    ASCII). If you need lower dimensional data, read the higher dimensional one
    and slice the data.

    :ivar x:           Access to 1D data along x-axis.
    :ivar y:           Access to 1D data along y-axis.
    :ivar z:           Access to 1D data along z-axis.
    :ivar xy:          Access to 2D data along xy-plane.
    :ivar xz:          Access to 2D data along xz-plane.
    :ivar yz:          Access to 2D data along yz-plane.
    :ivar xyz:         Access to 3D data.

    """

    # Usually we think in terms of dimensions xyz, but it is much more
    # convenint to index them with numbers. This dictionary provides a way
    # to go from one notation to the other. Internally, we always use the
    # index notation.
    _dim_indices = {
        "x": (0,),
        "y": (1,),
        "z": (2,),
        "xy": (0, 1),
        "xz": (0, 2),
        "yz": (1, 2),
        "xyz": (0, 1, 2),
    }

    def __init__(self, sd):
        """Constructor.

        :param sd: Simulation directory.
        :type sd: :py:class:`~.SimDir`
        """

        if not isinstance(sd, simdir.SimDir):
            raise TypeError("Input is not SimDir")

        # _all_griddata is a dictionary that maps dimension to an object
        # AllGridFunctions, which contains all the variables for which that
        # dimension is available
        self._all_griddata = {
            dim: AllGridFunctions(sd.allfiles, dim)
            for dim in self._dim_indices.values()
        }

    def _string_or_tuple_to_dimension_index(self, dimension):
        """Internally, we always refer to the different dimensions with their
        numerical index. However, it is more convenient to have public
        interfaces with xyz. This method takes a dimension that can be either
        a string or a tuple and returns the corresponding dimension in the
        index notation.

        E.g.: 'x' -> (0, ), or (1, 2) -> (1, 2), or 'xy' -> (0, 1)

        :returns:  tuple of dimensions, e.g. (0,1) for xy-plane.
        or string with the name, e.g. 'xy'
        """
        # If the input is a recognized tuple, just return it
        if dimension in self._dim_indices.values():
            return dimension

        # If the input is a recognized string, return the corresponding tuple
        if dimension in self._dim_indices.keys():
            return self._dim_indices[dimension]

        raise ValueError(f"{dimension} is not a recognized dimension")

    def __getitem__(self, dimension):
        """Get data with given dimensionality.

        :param dimension:  tuple of dimensions, e.g. (0,1) for xy-plane.
        or string with the name, e.g. 'xy'
        """
        return self._all_griddata[
            self._string_or_tuple_to_dimension_index(dimension)
        ]

    def __getattr__(self, attr):
        # This allows to call self.x, self.xy and so on
        if attr in self._dim_indices.keys():
            # We retrieve the data with __getitem__
            return self[attr]

        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {attr}"
        )

    def __contains__(self, dimension):
        return (
            self._string_or_tuple_to_dimension_index(dimension)
            in self._all_griddata
        )

    def __str__(self):
        """String representation"""
        return "\n".join([str(self[dim]) for dim in self._all_griddata])

    def total_filesize(self, unit="MB"):
        """Return the total size of the grid data files.
        Available units B, KB, MB and GB (in power of 1024 bytes).

        :param unit: Unit to use (in powers of 1024 bytes).
        :type unit: str among: ``B``, ``KB``, ``MB``, ``GB``.
        :returns: Total size of the grid data files.
        :rtype: float

        """
        # First we find all the unique files
        allfiles = set()
        for dim in self._all_griddata.keys():
            allfiles.update(self[dim].allfiles)
        return total_filesize(allfiles, unit=unit)



#!/usr/bin/env python3

# Copyright (C) 2020-2022 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/fourier_util.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.


"""The :py:mod:`~.frequencyseries` module provides a representation of
frequency series.

:py:class:`~.FrequencySeries` can be evenly or unevenly sampled, real or
complex. They support all the mathematical operations and operators you may
expect, and have additional methods, which include ones for taking derivatives,
integrals, apply windows, smooth the signal, take inverse Fourier transform, and
more. Most of these methods are available in two flavors: those that return a
new :py:class:`~.FrequencySeries`, and those which modify the object in place. The
latter have names with imperative verbs.

As in :py:class:`~.TimeSeries`, :py:class:`~.FrequencySeries` are derived from
the :py:class:`~.BaseSeries`, which in turn is derived from the abstract class
:py:class:`~.BaseNumerical`. Some of the capabilities of
:py:class:`~.FrequencySeries` (e.g., overloading the mathematical operators) are
implemented in the parent classes.

Additionally, two functions are defined in this module
:py:func:`~.load_FrequencySeries` and :py:func:`~.load_noise_curve`. Both loads
:py:class:`~.FrequencySeries` from a text file, but :py:func:`~.load_noise_curve`
is a simpler interface for real-valued signals.

"""

import numpy as np
from scipy.signal import argrelextrema

from kuibit import timeseries
from kuibit.series import BaseSeries, sample_common


def load_FrequencySeries(path, *args, complex_on_two_columns=False, **kwargs):
    """Load a text file as a :py:class:`~.FrequencySeries`.

    The backend is ``np.loadtxt``, and the unknown arguments passed to this
    function are given to ``np.loadtxt``. This can be used, for example, to
    specify the columns of the file that have to be read.

    :param path: Path of the file to be loaded.
    :type path: str
    :param complex_on_two_columns: When true, it is assumed that the real and
                                   the imaginary parts of the frequency series
                                   are on two columns. Otherwise, on one.
                                   This has to be False to load real data
                                   (e.g., noise curves).
    :type complex_on_two_columns: bool
    :returns: Loaded data as :py:class:`~.FrequencySeries`.
    :rtype: :py:class:`~.FrequencySeries`

    """
    if complex_on_two_columns:
        f, fft_real, fft_imag = np.loadtxt(
            path, unpack=True, ndmin=2, *args, **kwargs
        )
        fft = fft_real + 1j * fft_imag
    else:
        f, fft = np.loadtxt(path, unpack=True, ndmin=2, *args, **kwargs)
    return FrequencySeries(f, fft)


def load_noise_curve(path, *args, **kwargs):
    """Load a noise curve as a :py:class:`~.FrequencySeries`.

    Unknown arguments are passed to ``np.loadtxt``.

    This is syntactic sugar for the function :py:func:`~.load_FrequencySeries.`

    :param path: Path of the file to be loaded.
    :type path: str
    :returns: Noise curve.
    :rtype: :py:class:`~.FrequencySeries`
    """
    return load_FrequencySeries(
        path, complex_on_two_columns=False, *args, **kwargs
    )


class FrequencySeries(BaseSeries):
    """Class representing a Fourier spectrum.

    :ivar f:   Frequency
    :vartype f: 1D NumPy array or float

    :ivar fft:   Fourier transform
    :vartype fft: 1D NumPy array or float

    """

    def __init__(self, f, fft, guarantee_f_is_monotonic=False):
        """Create a :py:class:`~.FrequencySeries` providing frequencies and the value at
        those frequencies.

        It is your duty to make sure everything makes sense!

        When ``guarantee_f_is_monotonic`` is True, no checks will be perform to
        make sure that f is monotonically increasing (increasing performance).
        This should is used internally whenever a new series is returned from
        self (since we have already checked that ``f`` is good.) or in performance
        critical routines.

        :param f:  Frequencies.
        :type f: 1D NumPy array or float

        :param fft:   Fourier transform.
        :type fft: 1D NumPy array or float

        :param guarantee_f_is_monotonic: If true, it will be assumes that ``f`` is
                                         monotonically increasing.
        :type guarantee_f_is_monotonic: bool

        """
        # Use BaseClass init
        super().__init__(f, fft, guarantee_f_is_monotonic)

    # The following are the setters and getters, so that we can
    # resolve "self.f" and "self.fft"
    # Read documentation on BaseSeries
    @property
    def f(self):
        """Frequencies.

        :returns: Frequencies.
        :rtype: 1d NumPy array
        """
        # This is defined BaseClass
        return self.x

    @f.setter
    def f(self, f):
        # This is defined BaseClass
        self.x = f

    @property
    def fft(self):
        """Fourier components.

        :returns: Fourier components.
        :rtype: 1d NumPy array
        """
        # This is defined BaseClass
        return self.y

    @fft.setter
    def fft(self, fft):
        # This is defined BaseClass
        self.y = fft

    @property
    def fmin(self):
        """Return the minimum frequency.

        :returns:  Minimum frequency of the series.
        :rtype:    float
        """
        return self.xmin

    @property
    def fmax(self):
        """Return the maximum frequency.

        :returns:  Maximum frequency of the series.
        :rtype:    float
        """
        return self.xmax

    @property
    def frange(self):
        """Return the range of frequencies.
        The range is defined as the maximum frequency minus the minimum.

        :returns:  Range of the series (``f_max`` - ``f_min``).
        :rtype:    float
        """
        return self.fmax - self.fmin

    @property
    def amplitude(self):
        """Return the amplitude of frequencies.

        :returns:  Amplitude of the series.
        :rtype:    1d NumPy array
        """
        return abs(self.fft)

    # Writing amplitude all the times can be boring
    amp = amplitude

    @property
    def df(self):
        """Return the spacing (``delta_f``) if the series is regularly sampled,
        otherwise raise error.

        :returns: Frequency spacing (``delta_f``).
        :rtype: float

        """
        if not self.is_regularly_sampled():
            raise ValueError("Frequencyseries is not regularly sampled")

        return self.f[1] - self.f[0]

    def normalized(self):
        """Return a new :py:class:`~.FrequencySeries` with maximum amplitude of 1.

        :returns: Normalized :py:class:`~.FrequencySeries` series.
        :rtype: :py:class:`~.FrequencySeries`

        """
        m = self.amplitude.max()

        if m <= 0:
            raise ValueError("Non positive PSD maximum!")

        return self / m

    def normalize(self):
        """Scale values so that the maximum of the amplitude is 1."""
        self._apply_to_self(self.normalized)

    def low_passed(self, f):
        """Remove frequencies higher or equal than the given.

        :param f: Frequency above which the series will be zeroed.
        :type f: float
        :returns: Low-passed :py:class:`~.FrequencySeries`.
        :rtype: :py:class:`~.FrequencySeries`


        """
        msk = np.abs(self.f) <= f
        return FrequencySeries(self.f[msk], self.fft[msk])

    def low_pass(self, f):
        """Remove frequencies higher or equal than ``f`` (absolute value).

        :param f: Frequency above which series will be zeroed.
        :type f: float

        """
        self._apply_to_self(self.low_passed, f)

    def high_passed(self, f):
        """Remove frequencies lower or equal than the given.

        :param f: Frequency below which series will be zeroed.
        :type f: float
        :returns: High-passed :py:class:`~.FrequencySeries`.
        :rtype: :py:class:`~.FrequencySeries`

        """
        msk = np.abs(self.f) >= f
        return FrequencySeries(self.f[msk], self.fft[msk])

    def high_pass(self, f):
        """Remove all the frequencies smaller than f

        :param f: Frequency below which series will be zeroed.
        :type f: float

        """
        self._apply_to_self(self.high_passed, f)

    def negative_frequencies_removed(self):
        """Remove frequencies lower than 0.

        :returns:  Frequencyeseries with only positive frequencies.
        :rtype: :py:class:`~.FrequencySeries`

        """
        msk = self.f >= 0
        return FrequencySeries(self.f[msk], self.fft[msk])

    def negative_frequencies_remove(self):
        """Remove all the frequencies smaller than 0."""
        self._apply_to_self(self.negative_frequencies_removed)

    def band_passed(self, fmin, fmax):
        """Remove frequencies outside the given range ``fmin, fmax``.

        :param fmin: Minimum frequency.
        :type fmin: float
        :param fmax: Maximum frequency.
        :type fmax: float

        """
        ret = self.low_passed(fmax)
        return ret.high_passed(fmin)

    def band_pass(self, fmin, fmax):
        """Remove all the frequencies below ``fmin`` and above ``fmax``.

        :param fmin: Minimum frequency.
        :type fmin: float
        :param fmax: Maximum frequency.
        :type fmax: float

        """
        self._apply_to_self(self.band_passed, fmin, fmax)

    def peaks(self, amp_threshold=0.0):
        """Return the location and amplitude of the peaks of the amplitude.

        Peaks at the boundaries are not considered.

        :param amp_threshold: Ignore peaks smaller than this value.
        :type amp_threshold:  float

        :returns: Peaks of the amplitude as list of tuples of three elements:
                frequency of the bin in which the peak is, fitted frequency of
                the maximum with parabolic approximation, amplitude of the peak.
        :rtype: List of tuples of three elements.

        """
        peaks_f_indexes = argrelextrema(self.amp, np.greater)

        # We do not consider the peaks near the boundaries and those smaller
        # than amp_threshold
        peaks_f_indexes = [
            i
            for i in peaks_f_indexes[0]
            if 1 < i < len(self) - 2 and self.amp[i] > amp_threshold
        ]

        peaks_f = self.f[peaks_f_indexes]
        peaks_amp = self.amp[peaks_f_indexes]
        peaks_ff = [
            (
                self.f[i]
                + 0.5
                * self.df
                * (self.amp[i + 1] - self.amp[i - 1])
                / (2.0 * self.amp[i] - self.amp[i - 1] - self.amp[i + 1])
            )
            for i in peaks_f_indexes
        ]

        return tuple(zip(peaks_f, peaks_ff, peaks_amp))

    def peaks_frequencies(self, amp_threshold=0.0):
        """Return the frequencies of the peaks of the amplitude larger
        than ``amp_threshold``.

        The frequency is computed with a quadratic fit using the left and right
        neighbors.

        Peaks at the boundaries are not considered.

        :param amp_threshold: Ignore peaks smaller than this value.
        :type amp_threshold:  float

        :returns: Fitted frequencies of the peaks of the amplitude.
        :rtype: 1d NumPy array

        """
        return np.array([p[1] for p in self.peaks(amp_threshold)])

    def to_TimeSeries(self):
        """Perform a inverse Fourier transform.

        If only positive frequencies are found, we will assume that the
        original signal was real.

        :returns: Inverse Fourier transform.
        :rtype: :py:class:`.TimeSeries`

        """
        if self.is_masked():
            raise RuntimeError(
                "Fourier transform with masked data is not supported."
            )

        # If fmin >= 0, then, the signal was probably real to begin with.
        # We will restore the negative frequencies so that the operation
        # is the actual inverse of taking the dft.
        #
        if self.fmin < 0:

            # TimeSeries.to_FrequencySeries() rearranges the frequency so that
            # negative are on the left and positive on the right. Here, we undo
            # that.

            fft = np.fft.ifftshift(self.fft)

            t = np.fft.fftfreq(len(self.f), d=self.df)
            t = np.fft.fftshift(t)
            y = np.fft.ifft(fft)
        else:
            y = np.fft.irfft(self.fft)

            # To find the times we have to restore the negative frequencies
            # So, we simply recompute them assuming the current df
            frequencies = np.linspace(-self.fmax, self.fmax, len(y))
            t = np.fft.fftfreq(len(frequencies), d=self.df)
            # This not the order NumPy likes
            t = np.fft.fftshift(t)

        # We need the normalization df to compute physical quantities.
        # Intuitively, NumPy computes a_k = 1 / N \sum A_k exp(2 pi f t), to
        # transform this into an integral (true Fourier transform), we have to
        # multiply this by the measure of integration.

        # NOTE: Why exactly do we need len(t) here?
        #       It works, and PyCBC does it too. But what is the reason?
        return timeseries.TimeSeries(t, y * len(t) * self.df)

    def inner_product(
        self,
        other,
        fmin=0,
        fmax=np.inf,
        noises=None,
        same_domain=False,
    ):
        r"""Compute the (network) inner product with another :py:class:`~.FrequencySeries`.

        This is defined as:

        :math:`(h_1, h_2) = 4 \Re \int_{f_min}^{f_max} \frac{h_1 h_2^*}{S_n}`

        where ``S_n`` is the noise curve, and ``h_1``, ``h_2`` the series.

        In case multiple noise curves are supplied, compute

        :math:`(h_1, h_2) = \sum_{detectors}
        4 \Re \int_{f_min}^{f_max} \frac{h_1 h_2^*}{S_n}`

        This is the network inner product. To compute this quantity, you have
        to provide a list of noises.

        We assume that the :py:class:`~.FrequencySeries` are zero outside of the
        interval of definition, so if ``fmax`` (``fmin``) is larger (smaller)
        than the one available, it is effectively set to the one available.

        Since Fourier typically transforms explode at fmin = 0, the result of
        the integration is highly sensitive to regions near that frequency.

        If ``same_domain`` is True, it is assumed that all the
        :py:class:`~.FrequencySeries` involved are defined over the same
        frequencies. Turning this on speeds up computations, but it will result
        in incorrect results if the assumption is violated. If it is False, the
        domain of definition of the series is checked, if it is not the same for
        all the series, then they will be resampled.

        :param other: Second frequency series in the inner product.
        :type other: :py:class:`.FrequencySeries`
        :param fmin: Remove frequencies below this value.
        :type fmin: float
        :param fmax: Remove frequencies above this value.
        :type fmax: float
        :param noise: If None, no weight is applied.
        :type noise: :py:class:`.FrequencySeries`, list
                     of :py:class:`.FrequencySeries` or None
        :param same_domain: Whether to assume that the :py:class:`~.FrequencySeries`
                            are defined over the same frequencies. If you can
                            guarantee this, the computation will be faster.
        :type same_domain: bool

        :returns: Inner product between ``self`` and ``other``.
        :rtype: float

        """
        if not isinstance(other, type(self)):
            raise TypeError("The other object is not a FrequencySeries")

        if (
            (not isinstance(noises, type(self)))
            and (not isinstance(noises, list))
            and (noises is not None)
        ):
            raise TypeError("Noise is not (a list of) FrequencySeries or None")

        if fmin >= fmax:
            raise ValueError("fmin has to be smaller than fmax")

        if fmin < 0:
            raise ValueError("fmin has to be non-negative")

        if noises is None:
            # If noises is None, it means that the weight is one everywhere so,
            # we prepare a FrequencySeries that has the same frequencies as
            # self.
            # Everything will be resampled to a common set
            noises = FrequencySeries(self.f, np.ones_like(self.fft))

        # "res" = "resampled"
        to_be_res_list = [self, other]
        # Check if noises is a list, in that case add all the elements to
        # to to_be_res_list
        if isinstance(noises, list):
            to_be_res_list.extend(noises)
        else:
            # noises is not a list, just append it
            to_be_res_list.append(noises)

        if not same_domain:
            # Noises typically have Lorentian features, better to use a 0d
            # spline, so we enable piecewise_constant
            [res_self, res_other, *res_noises] = sample_common(
                to_be_res_list, resample=True, piecewise_constant=True
            )
        else:
            [res_self, res_other, *res_noises] = to_be_res_list

        for series in [res_self, res_other, *res_noises]:
            series.negative_frequencies_remove()
            series.band_pass(fmin=fmin, fmax=fmax)

        # Sum all the integrands
        integrand = FrequencySeries(res_self.f, np.zeros_like(res_self.fft))

        # We manipulate directly the fft fields because we have already
        # established that the series are defined on the same frequencies.
        # This is faster because it skips several sanity checks.
        for res_noise in res_noises:
            integrand += res_self * res_other.conjugate() / res_noise

        # 4 Re * \int
        # To align with PyCBC we do a rectangular integration here instead of
        # a trapeziodial one
        return 4 * np.sum(integrand.fft.real) * integrand.df

    def overlap(
        self,
        other,
        fmin=0,
        fmax=np.inf,
        noises=None,
        same_domain=False,
    ):
        r"""Compute the (network) overlap.

        This is defined as:

        :math:`\textrm{overlap} = (h_1, h_2) / \sqrt{(h_1, h_1)(h_2, h_2)}`

        where ``h_1``, ``h_2`` are the series.

        To compute is the network overlap, you have to provide a list of noises.

        We assume that the :py:class:`~.FrequencySeries` are zero outside of the
        interval of definition, so if ``fmax`` (``fmin``) is larger (smaller)
        than the one available, it is effectively set to the one available.

        Since Fourier typically transforms explode at fmin = 0, the result of
        the integration is highly sensitive to regions near that frequency.

        If ``same_domain`` is True, it is assumed that all the
        :py:class:`~.FrequencySeries` involved are defined over the same
        frequencies. Turning this on speeds up computations, but it will result
        in incorrect results if the assumption is violated. If it is False, the
        domain of definition of the series is checked, if it is not the same for
        all the series, then they will be resampled.


        :param other: Second frequency series in the overlap.
        :type other: :py:class:`.FrequencySeries`
        :param fmin: Remove frequencies below this value.
        :type fmin: float
        :param fmax: Remove frequencies above this value.
        :type fmax: float
        :param noise: If None, no weight is applied. If it is a list,
                      the netowrk overlap is computed.
        :type noise: (list of) :py:class:`.FrequencySeries` or None
        :param same_domain: Whether to assume that the :py:class:`~.FrequencySeries`
                            are defined over the same frequencies. If you can
                            guarantee this, the computation will be faster.
        :type same_domain: bool

        :returns: Overlap between ``self`` and ``other``.
        :rtype: float

        """
        # Error handling is done by inner_product
        inner_11 = self.inner_product(
            self,
            fmin=fmin,
            fmax=fmax,
            noises=noises,
            same_domain=same_domain,
        )
        inner_22 = other.inner_product(
            other,
            fmin=fmin,
            fmax=fmax,
            noises=noises,
            same_domain=same_domain,
        )
        inner_12 = self.inner_product(
            other, fmin, fmax, noises=noises, same_domain=same_domain
        )

        return inner_12 / np.sqrt(inner_11 * inner_22)



#!/usr/bin/env python3

# Copyright (C) 2020-2022 Gabriele Bozzola
#
# Based on by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/timeseries.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.timeseries` module provides a representation of time series
and convenience functions to create :py:class:`~.TimeSeries`.

:py:class:`~.TimeSeries` can be evenly or unevenly sampled are rich in features.
They support all the mathematical operations and operators you may expect, and
have additional methods, which include ones for taking derivatives, integrals,
apply windows, smooth the signal, take Fourier transform, and more. Most of
these methods are available in two flavors: those that return a new
:py:class:`~.TimeSeries`, and those which modify the object in place. The latter
have names with imperative verbs.

:py:class:`~.TimeSeries` are derived from the :py:class:`~.BaseSeries`, which in
turn is derived from the abstract class :py:class:`~.BaseNumerical`. Some of the
capabilities of :py:class:`~.TimeSeries` (e.g., overloading the mathematical
operators) are implemented in the parent classes.

The additional functions provided in :py:mod:`~.timeseries` are:

- :py:func:`~.remove_duplicated_iters` cleans the input arrays by removing duplicated times.
- :py:func:`~.unfold_phase` takes as argument a NumPy array representing a phase
  and unfolds it removing all the jumps of 2 pi. This is useful in gravitational
  wave analysis.
- :py:func:`~.combine_ts` takes a list of timeseries and removes all the overlapping segments.


"""

import warnings

import numpy as np
from scipy import signal

from kuibit import frequencyseries
from kuibit.series import BaseSeries


def remove_duplicated_iters(t, y):
    """Remove overlapping segments from a time series in (t,y).

    Only the latest of overlapping segments is kept, the rest
    removed.

    This function is used for cleaning up simulations with multiple
    checkpoints.

    Note, if t = [1, 2, 3, 4, 2, 3] the output will be [1, 2, 3].
    The '4' is discarded because it is not the last segment. The
    idea is that if this corresponds to a simulation restart, you
    may have changed the paramters, so that 4 is not anymore correct.
    We consider the second restart the "truth".

    :param t:  Times.
    :type t:   1D NumPy array
    :param y:  Values.
    :type t:   1D NumPy array

    :returns:  Strictly monotonic time series.
    :rtype:    :py:class:`~.TimeSeries`

    """
    # Let's unpack this code.
    # First, we define a new variable t2.
    #
    # t2 is essentially the "cumulative minimum" of the times of the time
    # series: t2[i] is the minimum up to index i
    #
    # To be more specific, we walk the time array backwards (t[::-1]) and
    # we compute the cumulative minima. Then, we reverse the array ([::-1])
    #
    # For example, if t = [1, 2, 3, 4, 2, 3]
    # Then t[::-1] = [3, 2, 4, 3, 2, 1], and
    # np.minimum.accumulate(t[::-1]) = [3, 2, 2, 2, 2, 1]
    # Reversing it: t2 = [1, 2, 2, 2, 2, 3]
    #
    # If t had no nuplicates t2 and t would be the same.
    # When t has duplicates, t2 is like t but in place of the duplicates
    # it has values that are equal or smaller.
    #
    # What we want is to have as output [1, 2, 3]
    # To get that, we compare t and t2. Values that are not duplicated
    # are those subtracted with the following are positive.
    # (t[:-1] < t2[1:])

    # First, we make sure that we are dealing with arrays and not lists
    t = np.array(t)
    y = np.array(y)

    t2 = np.minimum.accumulate(t[::-1])[::-1]
    # Here we append [True] because the last point is always included
    msk = np.hstack((t[:-1] < t2[1:], [True]))

    return TimeSeries(t[msk], y[msk])


def unfold_phase(phase):
    """Remove phase jumps to get a continuous (unfolded) phase.

    :param phase:     Phase wrapped around the provided jump.
    :type phase:      1D NumPy array

    :returns:         Phase plus multiples of pi chosen to minimize jumps.
    :rtype:           1D NumPy array
    """
    return np.unwrap(phase)


def combine_ts(series, prefer_late=True):
    """Combine several overlapping time series into one.

    In intervals covered by two or more time series, which data is used depends
    on the parameter prefer_late. If two segments start at the same time, the
    longer one gets used.

    :param series: The timeseries to combine.
    :type series:  list of :py:class:`~.TimeSeries`
    :param prefer_late: If true, prefer data that starts later for overlapping
                        segments, otherwise, use data from the ones that come
                        earlier.
    :type prfer_late:   bool

    :returns:      The combined time series
    :rtype:        :py:class:`~.TimeSeries`

    """

    # Late and early can be implemented in one shot by implementing one and
    # sending t -> -t for the other. For the "straight" way we implement
    # combine_ts_early.
    #
    # Let's consider a simple example for the reversed case
    # t1 = [1, 2, 3], t2 = [2, 3, 4], we want to have t = [1, 2, 3, 4]
    # sign = -1
    # timeseries = [t2, t1]
    # times = t2[::-1] = [4, 3, 2]
    # Next we walk through the remaining elements of the list
    # We want only to keep those with t < times[-1] = 2 (hence the switch)
    # In this case msk = [3, 2, 1] < 2 = [False, False, True], so
    # s_t[msk] = [1] and times = [4, 3, 2, 1].
    # At the end, we need to reverse the order again

    # sign is responsible of inverting the sorting key
    sign = -1 if prefer_late else 1

    # Tuples are compared lexicographically; the first items are compared; if
    # they are the same then the second items are compared, and so on.
    # So here we sort by tmin and tmax
    timeseries = sorted(series, key=lambda x: (sign * x.tmin, sign * x.tmax))
    # Now we are going to build up the t and y array, starting with the first
    times = timeseries[0].t[::sign]
    values = timeseries[0].y[::sign]
    for s in timeseries[1:]:
        # We need to walk backwards for "prefer_late"
        s_t = s.t[::sign]
        s_y = s.y[::sign]
        # We only keep those times that we don't have yet in the array times
        msk = s_t < times[-1] if prefer_late else s_t > times[-1]
        times = np.append(times, s_t[msk])
        values = np.append(values, s_y[msk])

    return TimeSeries(times[::sign], values[::sign])


class TimeSeries(BaseSeries):
    """This class represents real or complex valued time series.

    :py:class:`~.TimeSeries` are defined providing a time list or array and the
    corresponding values. For example,

    .. code-block:: python

        times = np.linspace(0, 2 * np.pi, 100)
        values = np.sin(times)

        ts = TimeSeries(times, values)


    Times cannot be empty or not monotonically increasing.
    Times and values must have the same length.

    TimeSeries are well-behaved classed, many operations and methods are
    implemented. For instance, you can sum/multiply two :py:class:`~.TimeSeries`.

    NumPy acts on TimeSeries cleanly, eg. ``np.log10(TimeSeries)`` is a
    :py:class:`~.TimeSeries` with ``log10(data)``.

    :py:class:`~.TimeSeries` have methods for smoothing, windowing, extracting phase and
    more.

    :ivar t: Times.
    :vartype t: 1D NumPy array or float
    :ivar y: Values.
    :vartype y: 1D NumPy array or float
    :ivar spline_real: Coefficients for a spline represent of the real part
                       of y.
    :vartype spline_real: tuple
    :ivar spline_imag: Coefficients for a spline represent of the real part
                       of y.
    :vartype spline_imag: tuple

    """

    # NOTE: Are you adding a function? Document it in timeseries.rst!

    def __init__(self, t, y, guarantee_t_is_monotonic=False):
        """Constructor.

        When guarantee_t_is_monotonic is True no checks will be perform to make
        sure that t is monotonically increasing (increasing performance). This
        should is used internally whenever a new series is returned from self
        (since we have already checked that t is good.) or in performance
        critical routines.

        :param t: Sampling times, need to be strictly increasing.
        :type t:  1D NumPy array or list

        :param y: Data samples, can be real or complex valued.
        :type y:  1D NumPy array or list

        :param guarantee_t_is_monotonic: The code will assume that t is
                                         monotonically increasing.
        :type guarantee_t_is_monotonic: bool

        """
        # Use BaseClass init
        super().__init__(t, y, guarantee_t_is_monotonic)

    # The following are the setters and getters, so that we can "resolve" .t
    # and .y

    # The @property decorator allows us to call .t instead of .t()
    @property
    def t(self):
        """Return the time.

        :returns: Times.
        :rtype: 1d NumPy array.
        """
        # This is defined BaseClass and it is where the actual data is stored.
        return self.x

    @t.setter
    def t(self, t):
        # This is defined BaseClass
        self.x = t

    @property
    def tmin(self):
        """Return the starting time.

        :returns:  Initial time of the timeseries.
        :rtype:    float
        """
        return self.xmin

    @property
    def tmax(self):
        """Return the final time.

        :returns:  Final time of the timeseries.
        :rtype:    float
        """
        return self.xmax

    @property
    def dt(self):
        """Return the timestep if the series is regularly sampled,
        otherwise raise error.

        :returns: Timestep of the series (if evenly sampled).
        :rtype: float

        """
        if not self.is_regularly_sampled():
            raise ValueError("Timeseries is not regularly sampled")

        return self.t[1] - self.t[0]

    @property
    def time_length(self):
        """Return the length of the covered time interval.

        :returns:  Length of time covered by the timeseries (tmax - tmin).
        :rtype:    float
        """
        return self.tmax - self.tmin

    duration = time_length

    def time_at_maximum(self, absolute=True):
        """Return the time at which the timeseries is maximum.

        :param absolute: Whether to take the absolute value of the
                         data.
        :type absolute: bool

        :returns:  Time at maximum. If ``absolute`` is True, then
                   time at absolute maximum.
        :rtype:    float
        """
        if absolute:
            return self.x_at_abs_maximum_y()
        return self.x_at_maximum_y()

    def time_at_minimum(self, absolute=True):
        """Return the time at which the timeseries is minimum.

        :param absolute: Whether to take the absolute value of the
                         data.
        :type absolute: bool

        :returns:  Time at minimum. If ``absolute`` is True, then
                   time at absolute minimum.
        :rtype:    float
        """
        if absolute:
            return self.x_at_abs_minimum_y()
        return self.x_at_minimum_y()

    def aligned_at_minimum(self):
        """Return a new timeseries with absolute minimum at t=0.

        :returns:  Timeseries shifted so that the minimum is a t=0.
        :rtype:    :py:class:`~.TimeSeries`
        """
        return self.time_shifted(-self.time_at_minimum())

    def align_at_minimum(self):
        """Time shift the series so that the absolute minimum is at t=0."""
        self._apply_to_self(self.aligned_at_minimum)

    def aligned_at_maximum(self):
        """Return a new timeseries with absolute maximum at t=0.

        :returns:  Timeseries shifted so that the maximum is a t=0.
        :rtype:    :py:class:`~.TimeSeries`
        """
        return self.time_shifted(-self.time_at_maximum())

    def align_at_maximum(self):
        """Time shift the series so that the absolute maximum is at t=0."""
        self._apply_to_self(self.aligned_at_maximum)

    def regular_resampled(self):
        """Return a new timeseries resampled to regularly spaced times,
        with the same number of points.

        :returns: Regularly resampled time series.
        :rtype:   :py:class:`~.TimeSeries`
        """
        t = np.linspace(self.tmin, self.tmax, len(self))
        return self.resampled(t)

    def regular_resample(self):
        """Resample the timeseries to regularly spaced times,
        with the same number of points.

        """
        self._apply_to_self(self.regular_resampled)

    def fixed_frequency_resampled(self, frequency):
        """Return a  :py:class:`~.TimeSeries` with same tmin and tmax
        but resampled at a fixed frequency. The final time will change
        if the frequency does not lead a integer number of timesteps.

        :param frequency: Sampling rate.
        :type frequency: float
        :returns:  Time series resampled with given frequency.
        :rtype:   :py:class:`~.TimeSeries`
        """
        dt = 1.0 / float(frequency)
        if dt > self.time_length:
            raise ValueError("Frequency too short for resampling")
        n = int(np.floor(self.time_length / dt))
        # We have to add one to n, so that we can include the tmax point
        new_times = self.tmin + np.arange(0, n + 1) * dt

        return self.resampled(new_times)

    def fixed_frequency_resample(self, frequency):
        """Resample the timeseries to regularly spaced times with the given frequency.
        The final time will change if the frequency does not lead a integer
        number of timesteps.

        :param frequency: Sampling rate.
        :type frequency: float

        """
        self._apply_to_self(self.fixed_frequency_resampled, frequency)

    def fixed_timestep_resample(self, timestep):
        """Resample the timeseries to regularly spaced times with given timestep.
        The final time will change if the timestep does not lead a integer
        number of timesteps.

        :param timestep: New timestep.
        :type timestep: float

        """
        self._apply_to_self(self.fixed_timestep_resampled, timestep)

    def fixed_timestep_resampled(self, timestep):
        """Return a new :py:class:`~.TimeSeries` with evenly spaced with the given
        timestep. The final time will change if the timestep does not lead a
        integer number of timesteps.

        :param timestep: New timestep.
        :type timestep: float
        :returns:  Time series resampled with given timestep.
        :rtype:   :py:class:`~.TimeSeries`

        """
        if timestep > self.time_length:
            raise ValueError("Timestep larger then duration of the TimeSeries")
        frequency = 1.0 / float(timestep)
        return self.fixed_frequency_resampled(frequency)

    def zero_padded(self, N):
        """Return a :py:class:`~.TimeSeries` that is zero-padded and that has
        in total ``N`` points.

        .. note::

            ``N`` is the final number of points, not the number of points added.

        This operation will work only if the series is equispaced.

        :param N: Total number of points of the output.
        :type N: int

        :returns: A new timeseries with in total N points where all
                  the trailing ones are zero.
        :rtype: :py:class:`~.TimeSeries`
        """
        N_new_zeros = N - len(self)

        if N_new_zeros < 0:
            raise ValueError(
                "Zero-padding cannot decrease the number of points"
            )

        new_zeros_t = np.linspace(
            self.tmax + self.dt,
            self.tmax + N_new_zeros * self.dt,
            N_new_zeros,
        )
        return TimeSeries(
            np.append(self.t, new_zeros_t),
            np.append(self.y, np.zeros(N_new_zeros)),
        )

    def zero_pad(self, N):
        """Pad the timeseries with zeros so that it has a total of N points.

        This operation will work only if the timeseries is equispaced and if N
        is larger than the number of points already present.

        .. note::

            ``N`` is the final number of points, not the number of points added.

        :param N: Total number new points with zeros at the end.
        :type N: int

        """
        self._apply_to_self(self.zero_padded, N)

    def mean_removed(self):
        """Return a :py:class:`~.TimeSeries` with mean removed, so that its new
        total average is zero.

        :returns: A new :py:class:`~.TimeSeries` with zero mean.
        :rtype: :py:class:`~.TimeSeries`
        """
        return TimeSeries(self.t, self.y - self.y.mean())

    def mean_remove(self):
        """Remove the mean value from the data."""
        self._apply_to_self(self.mean_removed)

    def initial_time_removed(self, time_init):
        """Return a :py:class:`~.TimeSeries` without the initial ``time_init`` amount of
        time.

        When ``tmin = 0``, this is the same as cropping, otherwise the
        difference is that in one case the time interval is specified, whereas
        in the other (cropping) the new ``tmin`` is specified.

        If a series goes from t=-1 to t=10 and you set time_init=2,
        the series will go from t=1 to t=10.

        :param time_init: Amount of time to be removed from the beginning.
        :type time_init: float

        :returns: A new :py:class:`~.TimeSeries` without the initial ``time_init``.
        :rtype: :py:class:`~.TimeSeries`

        """
        return self.cropped(init=self.tmin + time_init)

    def initial_time_remove(self, time_init):
        """Remove the first ``time_init`` amount of time in the timeseries.

        When ``tmin = 0``, this is the same as cropping, otherwise the
        difference is that in one case the time interval is specified, whereas
        in the other (cropping) the new ``tmin`` is specified.

        :param time_init: Amount of time to be removed from the beginning.
        :type time_init: float

        """
        self._apply_to_self(self.initial_time_removed, time_init)

    def final_time_removed(self, time_end):
        """Return a :py:class:`~.TimeSeries` without the final ``time_end`` amount of
        time.

        If a series goes from t=-1 to t=10 and you set time_end=2,
        the series will go from t=-1 to t=8.

        :param time_end: Amount of time to be removed from the end.
        :type time_end: float

        :returns: A new :py:class:`~.TimeSeries` without the final ``time_end``.
        :rtype: :py:class:`~.TimeSeries`

        """
        return self.cropped(end=self.tmax - time_end)

    def final_time_remove(self, time_end):
        """Remove the final ``time_end`` amount of time in the timeseries.

        :param time_end: Amount of time to be removed from the end.
        :type time_end: float

        """
        self._apply_to_self(self.final_time_removed, time_end)

    def time_shifted(self, tshift):
        """Return a new timeseries with time shifted by ``tshift`` so that
        what was t = 0 will be ``tshift``.

        :param tshift: Amount of time to shift.
        :type tshift: float
        :returns: A new :py:class:`~.TimeSeries` with time shifted.
        :rtype: :py:class:`~.TimeSeries`
        """
        return TimeSeries(self.t + tshift, self.y)

    def time_shift(self, tshift):
        """Shift the timeseries by ``tshift`` so that what was t = 0 will be ``tshift``.

        :param N: Amount of time to shift.
        :type N: float

        """
        self._apply_to_self(self.time_shifted, tshift)

    def phase_shifted(self, pshift):
        """Return a new :py:class:`~.TimeSeries` with complex phase shifted by ``pshift``.
        If the signal is real, it is turned complex with phase of ``pshift``.

        :param pshift: Amount of phase to shift.
        :type pshift: float
        :returns: A new :py:class:`~.TimeSeries` with phase shifted.
        :rtype: :py:class:`~.TimeSeries`

        """
        return TimeSeries(self.t, self.y * np.exp(1j * pshift))

    def phase_shift(self, pshift):
        """Shift the complex phase timeseries by ``pshift``. If the signal is real,
        it is turned complex with phase of ``pshift``.

        :param pshift: Amount of phase to shift.
        :type pshift: float

        """
        self._apply_to_self(self.phase_shifted, pshift)

    def time_unit_changed(self, unit, inverse=False):
        """Return a new :py:class:`~.TimeSeries` with time scaled by ``unit``.

        This amounts to sending t to ``t / unit``. For example, if initially the
        units where seconds, with unit=1e-3 the new units will be milliseconds.

        When inverse is True, the opposite is done and t is sent to ``t *
        unit``. This is useful to convert geometrized units to physical units
        with :py:mod:`~.unitconv`. For example,

        .. code-block:: python

            # Gravitational waves in geometrized units
            gw_cu = TimeSeries(...)
            # Gravitational waves in seconds, assuming a mass of 1 M_sun
            CU = uc.geom_umass_msun(1)
            gw_s = gw_cu.time_unit_changed(CU.time, inverse=True)

        :param unit: New time unit.
        :type unit: float
        :param inverse: If True, time = 1 -> time = unit, otherwise
                        time = unit -> 1.
        :type inverse: bool

        :returns: A :py:class:`~.TimeSeries` with new time unit.
        :rtype: :py:class:`~.TimeSeries`

        """
        factor = unit if inverse else 1 / unit
        return TimeSeries(self.t * factor, self.y)

    def time_unit_change(self, unit, inverse=False):
        """Rescale time units by unit.

        This amounts to sending t to ``t / unit``. For example, if initially the
        units where seconds, with unit=1e-3 the new units will be milliseconds.

        When inverse is True, the opposite is done and t is sent to ``t *
        unit``. This is useful to convert geometrized units to physical units
        with :py:mod:`~.unitconv`. For example,

        .. code-block:: python

            # Gravitational waves in geometrized units
            gw_cu = TimeSeries(...)
            # Gravitational waves in seconds, assuming a mass of 1 M_sun
            CU = uc.geom_umass_msun(1)
            gw_s = gw_cu.time_unit_changed(CU.time, inverse=True)

        :param unit: New time unit.
        :type unit: float
        :param inverse: If True, time = 1 -> time = unit, otherwise
                        time = unit -> 1.
        :type inverse: bool
        """
        self._apply_to_self(self.time_unit_changed, unit, inverse)

    def redshifted(self, z):
        """Return a new :py:class:`~.TimeSeries` with time rescaled so that frequencies
        are redshifted by ``1 + z``.

        :param z: Redshift factor.
        :type z: float

        :returns: A new redshifted :py:class:`~.TimeSeries`.
        :rtype: :py:class:`~.TimeSeries`

        """
        return self.time_unit_changed(1 + z, inverse=True)

    def redshift(self, z):
        """Apply redshift to the data by rescaling the time so that the frequencies
        are redshifted by ``1 + z``.

        :param z: Redshift factor.
        :type z: float

        """
        self._apply_to_self(self.redshifted, z)

    def unfolded_phase(self, t_of_zero_phase=None):
        """Compute the complex phase of a complex-valued signal such that no phase
        wrap-around occur, i.e. if the input is continuous, so is the output.
        Optionally, add a phase shift such that phase is zero at the given time.

        :param t_of_zero_phase: Time at which the phase is set to zero.
        :type t_of_zero_phase:   float or None

        :returns:   Continuous complex phase.
        :rtype:     :py:class:`~.TimeSeries`

        """
        ret = TimeSeries(self.t, unfold_phase(np.angle(self.y)))
        if t_of_zero_phase is not None:
            ret -= ret(t_of_zero_phase)
        return ret

    def phase_angular_velocity(self, use_splines=True, tsmooth=None, order=3):
        """Compute the phase angular velocity, i.e. the time derivative of the
        complex phase.

        Optionally smooth the with a savgol filter with smoothing length
        tsmooth and order order. If you do so, the timeseries is resampled to
        regular timesteps.

        :param use_splines: Wheter to use splines of finite differencing for
                            the derivative.
        :type use_splines: bool
        :param tsmooth: Time over which smoothing is applied.
        :type tsmooth: float
        :param order: Order of the for the savgol smoothing.
        :type order: int

        :returns:  Time derivative of the complex phase.
        :rtype:    :py:class:`~.TimeSeries`
        """
        if use_splines:
            ret_value = self.unfolded_phase().spline_differentiated()
        else:
            ret_value = self.unfolded_phase().differentiated()

        if tsmooth is not None:
            ret_value.savgol_smooth_time(tsmooth, order)

        return ret_value

    def phase_frequency(self, use_splines=True, tsmooth=None, order=3):
        """Compute the phase frequency, i.e. the time derivative
        of the complex phase divided by 2 pi.

        Optionally smooth the with a savgol filter with smoothing length
        tsmooth and order order. If you do so, the timeseries is resampled
        to regular timesteps.

        :param use_splines: Wheter to use splines of finite differencing for
                            the derivative.
        :type use_splines: bool
        :param tsmooth: Time over which smoothing is applied.
        :type tsmooth: float
        :param order: Order of the for the savgol smoothing.
        :type order: int

        :returns:  Time derivative of the complex phase divided by 2 pi
        :rtype:    :py:class:`~.TimeSeries`
        """
        return self.phase_angular_velocity(use_splines, tsmooth, order) / (
            2 * np.pi
        )

    def windowed(self, window_function, *args, **kwargs):
        """Return a :py:class:`~.TimeSeries` windowed with ``window_function``.

        ``window_function`` has to be a function that takes as first argument
        the number of points of the signal. ``window_function`` can take
        additional arguments as passed by ``windowed``. Alternatively,
        ``window_function`` can be a string with the name of the window
        function, if this is already implemented in :py:class:`~.TimeSeries`
        (e.g., ``tukey``).

        :param window_function: Window function to apply to the timeseries.
        :type window_function: callable or str

        :returns:  New windowed :py:class:`~.TimeSeries`.
        :rtype:    :py:class:`~.TimeSeries`

        """
        if not self.is_regularly_sampled():
            warnings.warn(
                "TimeSeries is not regularly samples. Resampling.",
                RuntimeWarning,
            )
            regular_ts = self.regular_resampled()
        else:
            regular_ts = self

        if callable(window_function):
            window_array = window_function(len(regular_ts), *args, **kwargs)
            return TimeSeries(regular_ts.t, regular_ts.y * window_array)

        if isinstance(window_function, str):
            window_function_method = f"{window_function}_windowed"
            if not hasattr(self, window_function_method):
                raise ValueError(f"Window {window_function} not implemented")
            window_function_callable = getattr(self, window_function_method)
            return window_function_callable(*args, **kwargs)

        raise TypeError("Window function is neither a callable or a string")

    def window(self, window_function, *args, **kwargs):
        """Apply window_function to the data.

        ``window_function`` has to be a function that takes as first argument
        the number of points of the signal. ``window_function`` can take
        additional arguments as passed by ``windowed``. Alternatively,
        ``window_function`` can be a string with the name of the window
        function, if this is already implemented in :py:class:`~.TimeSeries`
        (e.g., ``tukey``).

        :param window_function: Window function to apply to the timeseries.
        :type window_function: callable or str

        """
        self._apply_to_self(self.windowed, window_function, *args, **kwargs)

    def tukey_windowed(self, alpha):
        """Return a :py:class:`~.TimeSeries` with Tukey window with parameter ``alpha``
        applied.

        :param alpha: Tukey parameter.
        :type alpha: float

        :returns:  New windowed :py:class:`~.TimeSeries`.
        :rtype:    :py:class:`~.TimeSeries`

        """
        return self.windowed(signal.tukey, alpha)

    def tukey_window(self, alpha):
        """Apply Tukey window with parameter ``alpha``.

        :param alpha: Tukey parameter.
        :type alpha: float

        """
        self.window(signal.tukey, alpha)

    def hamming_windowed(self):
        """Return a timeseries with Hamming window applied.

        :returns:  New windowed :py:class:`~.TimeSeries`.
        :rtype:    :py:class:`~.TimeSeries`

        """
        return self.windowed(signal.hamming)

    def hamming_window(self):
        """Apply Hamming window."""
        self.window(signal.hamming)

    def blackman_windowed(self):
        """Return a timeseries with Blackman window applied."""
        return self.windowed(signal.blackman)

    def blackman_window(self):
        """Apply Blackman window."""
        self.window(signal.blackman)

    def savgol_smoothed_time(self, tsmooth, order=3):
        """Return a resampled timeseries with uniform timesteps, smoothed with
        ``savgol_smooth`` with a window that is ``tsmooth`` in time (as opposed
        to a number of points).

        :param tsmooth: Time interval over which to smooth.
        :type tsmooth: float
        :param order: Order of the filter.
        :type order: int

        :returns:  New smoothed and resampled :py:class:`~.TimeSeries`.
        :rtype:    :py:class:`~.TimeSeries`

        """
        if not self.is_regularly_sampled():
            warnings.warn(
                "TimeSeries is not regularly samples. Resampling.",
                RuntimeWarning,
            )
            ts = self.regular_resampled()
        else:
            ts = self
        dt = ts.t[1] - ts.t[0]
        # The savgol method requires a odd window
        # If it is not, we add one point
        window = int(np.rint(tsmooth / dt))
        window = window + 1 if (window % 2 == 0) else window
        return self.savgol_smoothed(window, order)

    def savgol_smooth_time(self, tsmooth, order=3):
        """Resample the timeseries with uniform timesteps, smooth it with
        ``savgol_smooth`` with a window that is ``tsmooth`` in time (as opposed to a
        number of points).

        :param tsmooth: Time interval over which to smooth.
        :type tsmooth: float
        :param order: Order of the filter.
        :type order: int

        """
        self._apply_to_self(self.savgol_smoothed_time, tsmooth, order)

    def to_FrequencySeries(self):
        """Return a :py:class:`~.FrequencySeries` that is the Fourier transform of the
        timeseries.

        If the signal is not complex, only positive frequencies are kept.

        If the timeseries is not regularly sampled, it will be resampled before
        transforming.

        :: warning:

            To have meaningful results, you should consider removing the
            mean and windowing the signal before calling this method!

        :returns: Fourier Transform.
        :rtype: :py:class:`~.FrequencySeries`

        """
        if self.is_masked():
            raise RuntimeError(
                "Fourier transform with masked data is not supported."
            )

        if not self.is_regularly_sampled():
            warnings.warn(
                "TimeSeries is not regularly samples. Resampling.",
                RuntimeWarning,
            )
            regular_ts = self.regular_resampled()
        else:
            regular_ts = self

        dt = regular_ts.dt

        if self.is_complex():
            frequencies = np.fft.fftfreq(len(regular_ts), d=dt)
            fft = np.fft.fft(regular_ts.y)

            f = np.fft.fftshift(frequencies)
            fft = np.fft.fftshift(fft)
        else:
            # Note the "r"
            f = np.fft.rfftfreq(len(regular_ts), d=dt)
            fft = np.fft.rfft(regular_ts.y)

        # We need the normalization dt to compute physical quantities.
        # Intuitively, NumPy computes A_k = \sum a_k exp(-2 pi f t), to
        # transform this into an integral (true Fourier transform), we have to
        # multiply this by the measure of integration.
        return frequencyseries.FrequencySeries(f, fft * dt)



#!/usr/bin/env python3

# Copyright (C) 2020-2022 Gabriele Bozzola
# sYlm: Copyright (C) Christian Reisswig
# ra_dec_to_theta_phi: Copyright (C) 2020-2022 Yuk Tung Liu
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.


"""The module :py:mod:`~.gw_utils` contains convenience functions and
structures to analyze and work with gravitational waves.

First, the ``Detectors`` object is defined. ``Detectors`` is a named tuple
with fields "hanford", "livingston", "virgo". This is used every time we
deal with specific detectors.

The functions provided are:

- :py:func:`~.luminosity_distance_to_redshift`: convert a given luminosity
  distance to a redshift in the ΛCDM cosmology.
- :py:func:`~.sYlm`: return the spin-weighted spherical harmonics at a given
  angle.
- :py:func:`~.ra_dec_to_theta_phi`: convert right ascension and declination to
  spherical coordinates.
- :py:func:`~.antenna_responses`: compute the antenna responses of a given
  angle.
- :py:func:`~.antenna_responses_from_sky_localization`: compute the antenna
  responses for known detectors at a given sky localization.

- :py:func:`~.signal_to_noise_ratio_from_strain`: compute the signal to noise
  for a given signal and a given noise curve.

"""

import datetime
from collections import namedtuple

import numpy as np
from scipy import integrate, optimize

import kuibit.timeseries as ts
import kuibit.unitconv as uc

# This is just a convenience to avoid having to remember the order of
# the output (and for easy of extension)
# One can access the output as detectos.hanford, or
# detectors['hanford']
Detectors = namedtuple("Detectors", "hanford livingston virgo")


def luminosity_distance_to_redshift(
    luminosity_distance,
    Omega_m=0.309,
    Omega_L=0.691,
    initial_guess=0.1,
):
    r"""Compute redshift from luminosity distance in Mpc assuming the ΛCDM cosmology.

    This function is useful to correctly reproduce observed signals
    from cosmological sources (e.g., binary black holes far away).

    The redshift is computed via root-finding, so an initial guess is needed.

    :param luminosity_distance: Luminosity distance in megaparsec.
    :type luminosity_distance: float
    :param Omega_m: :math:`\Omega_m` (matter) cosmological parameter.
    :type Omega_m: float
    :param Omega_L: :math:`\Omega_m` (dark energy) cosmological parameter.
    :type Omega_L: float
    :param initial_guess: Initial guess to the redshift for the
                          root-finding routine.
    :type initial_guess: float

    :returns z: Redshift
    :rtype z: float

    """
    distance_in_m = luminosity_distance * uc.MEGAPARSEC_SI  # m
    H0 = uc.H0_SI  # 1/s
    c = uc.C_SI  # m/s

    def function_to_root_find(z):
        def z_to_DL(z):
            def DL_integral(z):
                return 1 / np.sqrt(Omega_m * (1 + z) ** 3 + Omega_L)

            return c / H0 * (1 + z) * integrate.quad(DL_integral, 0, z)[0]

        return np.abs(distance_in_m - z_to_DL(z))

    redshift = optimize.root(function_to_root_find, initial_guess)

    if redshift["status"] != 1:
        raise RuntimeError("Conversion between distance and redshift failed!")

    return redshift["x"][0]


def sYlm(ss, ll, mm, theta, phi):
    """Compute spin-weighted spherical harmonics at the angles ``theta`` and ``phi``.

    When ``ss = 0``, these are spherical harmonics.

    :param ss: Spin weight.
    :type ss: int
    :param ll: :math:`l` multipolar number.
    :type ll: int
    :param mm: :math:`m` multipolar number.
    :type mm: int
    :param theta: Meridional angle.
    :type theta: float
    :param phi: Azimuthal angle.
    :type phi: float

    :returns sYlm: Spin-weighted spherical harmonic evaluated at
                  ``theta`` and ``phi``.
    :rtype sYlm: float

    """
    # Code by Christian Reisswig

    # Recursion function for spin-weighted spherical harmonics
    def s_lambda_lm(local_s, local_l, local_m, x):
        # Coefficient function for spin-weighted spherical harmonics
        def sYlm_Cslm(local_s, local_l, local_m):
            return np.sqrt(
                local_l
                * local_l
                * (4.0 * local_l * local_l - 1.0)
                / (
                    (local_l * local_l - local_m * local_m)
                    * (local_l * local_l - local_s * local_s)
                )
            )

        Pm = np.power(-0.5, local_m)

        if local_m != local_s:
            Pm = Pm * np.power(1.0 + x, (local_m - local_s) * 1.0 / 2)
        if local_m != -local_s:
            Pm = Pm * np.power(1.0 - x, (local_m + local_s) * 1.0 / 2)

        Pm = Pm * np.sqrt(
            np.math.factorial(2 * local_m + 1)
            * 1.0
            / (
                4.0
                * np.pi
                * np.math.factorial(local_m + local_s)
                * np.math.factorial(local_m - local_s)
            )
        )

        if local_l == local_m:
            return Pm

        Pm1 = (
            (x + local_s * 1.0 / (local_m + 1))
            * sYlm_Cslm(local_s, local_m + 1, local_m)
            * Pm
        )

        if local_l == local_m + 1:
            return Pm1

        for n in range(local_m + 2, local_l + 1):
            Pn = (x + local_s * local_m * 1.0 / (n * (n - 1.0))) * sYlm_Cslm(
                local_s, n, local_m
            ) * Pm1 - sYlm_Cslm(local_s, n, local_m) * 1.0 / sYlm_Cslm(
                local_s, n - 1, local_m
            ) * Pm
            Pm = Pm1
            Pm1 = Pn

        return Pn

    Pm = 1.0

    mult_l = ll
    mult_m = mm
    mult_s = ss

    if mult_l < 0:
        return 0
    if abs(mult_m) > mult_l or mult_l < abs(mult_s):
        return 0

    if abs(mm) < abs(ss):
        mult_s = mm
        mult_m = ss
        if (mult_m + mult_s) % 2:
            Pm = -Pm

    if mult_m < 0:
        mult_s = -mult_s
        mult_m = -mult_m
        if (mult_m + mult_s) % 2:
            Pm = -Pm

    result = Pm * s_lambda_lm(mult_s, mult_l, mult_m, np.cos(theta))

    return complex(result * np.cos(mm * phi), result * np.sin(mm * phi))


def ra_dec_to_theta_phi(right_ascension, declination, time_utc):
    """Compute the spherical angles ``theta`` and ``phi`` for Hanford, Livingston
    and Virgo for a given source localization.

    ``utc_time`` has to have the following formatting: ``%Y-%m-%d %H:%M``,
    (eg ``2015-09-14 09:50:45``)

    :param right_ascension: Right ascension of the source in degrees.
    :type right_ascension: float
    :param declination: Declination of the source in degrees.
    :type declination: float
    :param time_utc: UTC time of the event.
    :type declination: str

    :returns spherical coordinates: ``Theta``, ``phi`` for the different detectors.
    :rtype: namedtuple with fields hanford, livingston, and virgo

    """
    # NOTE: This function can make use of some cleanup and double-checking
    # This code is based on the one written by Yuk Tung Liu.

    # RA and DEC of the GW source:
    alpha_hours = right_ascension
    delta_degrees = declination

    # degrees to radian
    deg_to_rad = np.pi / 180.0

    alpha = alpha_hours * 15.0 * deg_to_rad
    delta = delta_degrees * deg_to_rad

    # LIGO detector geometry from
    # http://www.ligo.org/scientists/GW100916/GW100916-geometry.html
    # and also
    # http://www.ligo.org/scientists/GRB051103/GRB051103-geometry.php
    # The two pages are consistent.

    # LIGO Hanford (H) detector geometry
    # Latitude, Longitude and azimuth of the x-arm
    # Note: Longitude is measured positively westward from Greenwich
    # Azimuth is measured from the South point, turning positive to the West.
    lat_H = (46 + 27.0 / 60 + 19.0 / 3600) * deg_to_rad
    long_H = (119 + 24.0 / 60 + 28.0 / 3600) * deg_to_rad
    xazi_H = (180 - 36) * deg_to_rad

    # LIGO Livingston (L) detector geometry
    # Latitude, Longitude and azimuth of the x-arm
    lat_L = (30 + 33.0 / 60 + 46.0 / 3600) * deg_to_rad
    long_L = (90 + 46.0 / 60 + 27.0 / 3600) * deg_to_rad
    xazi_L = (90 - 18) * deg_to_rad

    # Virgo detector geometry
    # https://arxiv.org/pdf/1706.09505.pdf (Table 3.1)
    lat_V = (43 + 37.0 / 60 + 53.0 / 3600) * deg_to_rad
    long_V = -(10 + 30.0 / 60 + 16.0 / 3600) * deg_to_rad
    xazi_V = -(180 - 19) * deg_to_rad

    # Time of detection of GW150914: 2015-09-14 09:50:45 UTC
    # The Greenwich sidereal time theta_G is calculated by
    # the formula at https://en.wikipedia.org/wiki/Sidereal_time

    base_date = datetime.datetime.strptime(
        "2000-01-01 12:00:00", "%Y-%m-%d %H:%M:%S"
    )
    date = datetime.datetime.strptime(time_utc, "%Y-%m-%d %H:%M:%S")
    # Days between DATE and 2000-01-01 12:00
    D = (date - base_date).total_seconds() / 86400
    theta_G = 18.697374558 + 24.06570982441908 * D
    # theta_G mod 24h
    theta_G = theta_G - np.floor(theta_G / 24.0) * 24
    theta_G = theta_G * 15 * deg_to_rad

    # Equatorial -> horizontal using the formulae at
    # https://en.wikipedia.org/wiki/Celestial_coordinate_system

    # Hour angles
    h_H = theta_G - long_H - alpha
    h_L = theta_G - long_L - alpha
    h_V = theta_G - long_V - alpha

    # Zenith distance theta and azimuth A
    theta_H = np.arccos(
        np.sin(lat_H) * np.sin(delta)
        + np.cos(lat_H) * np.cos(delta) * np.cos(h_H)
    )
    theta_L = np.arccos(
        np.sin(lat_L) * np.sin(delta)
        + np.cos(lat_L) * np.cos(delta) * np.cos(h_L)
    )
    theta_V = np.arccos(
        np.sin(lat_V) * np.sin(delta)
        + np.cos(lat_V) * np.cos(delta) * np.cos(h_V)
    )

    A_H = np.arctan2(
        np.cos(delta) * np.sin(h_H),
        np.cos(delta) * np.cos(h_H) * np.sin(lat_H)
        - np.sin(delta) * np.cos(lat_H),
    )
    phi_H = xazi_H - A_H

    A_L = np.arctan2(
        np.cos(delta) * np.sin(h_L),
        np.cos(delta) * np.cos(h_L) * np.sin(lat_L)
        - np.sin(delta) * np.cos(lat_L),
    )
    phi_L = xazi_L - A_L

    A_V = np.arctan2(
        np.cos(delta) * np.sin(h_V),
        np.cos(delta) * np.cos(h_V) * np.sin(lat_V)
        - np.sin(delta) * np.cos(lat_V),
    )
    phi_V = xazi_V - A_V

    coords = Detectors(
        hanford=(theta_H, phi_H),
        livingston=(theta_L, phi_L),
        virgo=(theta_V, phi_V),
    )

    return coords


def antenna_responses(theta, phi, polarization=0):
    """Return the antenna response pattern of a detector on the z = 0 plane
    with the arms on the x and y directions for a given localization defined
    by the spherical angles ``theta`` and ``phi``.

    :param theta: Meridional angle.
    :type theta: float
    :param phi: Azimuthal angle.
    :type phi: float
    :param polarization: Polarization angle of the wave.
    :type polarization: float

    :returns: Antenna response for cross and plus polarizations (in this order).
    :rtype: tuple of floats.

    """
    # http://research.physics.illinois.edu/cta/movies/bhbh_sim/wavestrain.html

    Fp = 0.5 * (1 + np.cos(theta) * np.cos(theta)) * np.cos(2 * phi) * np.cos(
        2 * polarization
    ) - np.cos(theta) * np.sin(2 * phi) * np.sin(2 * polarization)
    Fc = 0.5 * (1 + np.cos(theta) * np.cos(theta)) * np.cos(2 * phi) * np.sin(
        2 * polarization
    ) + np.cos(theta) * np.sin(2 * phi) * np.cos(2 * polarization)

    return (Fc, Fp)


def antenna_responses_from_sky_localization(
    right_ascension, declination, time_utc, polarization=0
):
    """Return the antenna responses for Hanford, Livingston and Virgo for a
    given source.

    See,
    http://research.physics.illinois.edu/cta/movies/bhbh_sim/wavestrain.html.

    ``utc_time`` has to have the following formatting: ``%Y-%m-%d %H:%M``,
    (eg ``2015-09-14 09:50:45``)

    :param right_ascension: Right ascension of the source in degrees
    :type right_ascension: float
    :param declination: Declination of the source in degrees
    :type declination: float
    :param time_utc: UTC time of the event
    :type declination: str
    :param polarization: Polarization of the wave
    :type polarization: float

    :returns antenna_pattern: Cross and plus antenna pattern for the different
                             interferometers.
    :rtype: namedtuple with fields hanford, livingston, and virgo

    """

    coords = ra_dec_to_theta_phi(right_ascension, declination, time_utc)

    Fc_H, Fp_H = antenna_responses(*coords.hanford, polarization)
    Fc_L, Fp_L = antenna_responses(*coords.livingston, polarization)
    Fc_V, Fp_V = antenna_responses(*coords.virgo, polarization)

    antenna = Detectors(
        hanford=(Fc_H, Fp_H),
        livingston=(Fc_L, Fp_L),
        virgo=(Fc_V, Fp_V),
    )

    return antenna


# Extrapolation of GWs at infinity.
# Follows what done in the NRAR collaboration (1307.5307)
# This is what we are going to do:

# 1. Assume that the spacetime is almost Schwarzschild with mass M.
# 2. Choose a set of retarded times u_i where GWs have to be evaluated
# 3. Compute the coordinate times t_i that correspond to the retarded
#    times u_i at the radius r. This uses tortoise coordinates.
# 4. Interpolate the waveforms at the coordinate times corresponding to
#    the retarded times u_i for different extraction radii.
#    Now our waves are so that they are evaluate at different t_i but at
#    the same u_i.
# 5. We do this process for a bunch of extraction radii (rex1, rex2, rex3, ...)
#    So, we should have wave1, wave2, wave3, with all the same number of
#    points that correspond to the same retarded time.
# 6. For each element in u_i, fit all the waves (wave1, wave2, ...) with a
#    polynomial of the form a_n/r^n.


def Schwarzschild_radius_to_tortoise(radii, mass):
    """Transform radial coordinates ``radii`` to tortoise coordinates assuming mass
    ``mass``.

    Equation (26) in 1307.5307.

    :param radii: Radius in Schwarzschild coordinates.
    :type radii: float or 1D NumPy array
    :param mass: ADM mass.
    :type mass: float

    :returns: Tortoise radii.
    :rtype: float or 1D NumPy array

    """
    return radii + 2 * mass * np.log(radii / (2 * mass) - 1)


def retarded_times_to_coordinate_times(retarded_times, radii, mass):
    """Compute the coordinate times corresponding to the retarded times at the
      coordinate radii.

    First, the tortoise radius is computed from ``radii``, then the
    coordinate times are computed with :math:`t = u + r_{tortoise}(radii)`.

    This function is used to extrapolate gravitational waves to infinity.

    :param retarded_times: Retarded times.
    :type retarded_times: float or 1D NumPy array
    :param radii: Radii of evaluation.
    :type radii: float or 1D NumPy array
    :param mass: ADM mass (needed to compute tortoise radius).
    :type mass: float

    :returns: Coordinate times corresponding to the given retarded times
              and evaluation radii.
    :rtype: float or 1D NumPy array

    """
    return retarded_times + Schwarzschild_radius_to_tortoise(radii, mass)


def _coordinate_times_to_retarded_times(coordinate_times, radii, mass):
    """Compute the coordinate times corresponding to the retarded times
      at the given coordinate radii.

    This function is used to extrapolate gravitational waves to infinity.

    :param retarded_times: Coordinate times.
    :type retarded_times: float or 1D NumPy array
    :param radii: Radii (it can be just one)
    :type radii: float or 1D NumPy array
    :param mass: ADM mass
    :type mass: float

    :returns: Retarded times corresponding to the given coordinate times
              and evaluation radii.
    :rtype: float or 1D NumPy array

    """
    return coordinate_times - Schwarzschild_radius_to_tortoise(radii, mass)


def signal_to_noise_ratio_from_strain(
    h, *args, noise=None, fmin=0, fmax=np.inf, window_function=None, **kwargs
):
    r"""Return the signal to noise ratio given a strain and a power spectal density
    distribution for a detector.

    If ``window_function`` is not None, the window will be applied to the signal.
    All the unknown arguments are passed to the window function.

    The SNR is computed as :math:`sqrt of 4 \int_fmin^fmax |\tilde{h} f|^2 / Sn(f) d f`
    (equation from 1408.0740)

    :param h: Gravitational-wave strain.
    :type h: :py:class:`~.TimeSeries`
    :param noise: Power spectral density of the noise of the detector.
    :type noise: :py:class:`~.FrequencySeries`
    :param fmin: Minimum frequency over which to compute the SNR.
    :type fmin: float
    :param fmax: Maximum frequency over which to compute the SNR.
    :type fmax: float
    :param window_function: If not None, apply ``window_function`` to the
                            series before computing the strain.
    :type window_function: callable, str, or None
    :param args, kwargs: All the additional parameters are passed to
                         the window function.

    :returns: Signal-to-noise ratio.
    :rtype: float

    """
    if not isinstance(h, ts.TimeSeries):
        raise TypeError("Strain has to be a TimeSeries")
    # First, we window to avoid spectral leakage
    h_win = h.windowed(window_function, *args, **kwargs)
    # Then, we take the Fourier transform
    h_fft = h_win.to_FrequencySeries()
    # The SNR is just the inner product of h_fft with itself
    return np.sqrt(
        h_fft.inner_product(h_fft, noises=noise, fmin=fmin, fmax=fmax)
    )



#!/usr/bin/env python3

# Copyright (C) 2020-2022 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""kuibit provides utilities to post-process simulations performed with the
Einstein Toolkit (or any Cactus-based code).

"""

__version__ = "1.3.5"

__bibtex__ = """\
@article{kuibit,
       author = {{Bozzola}, Gabriele},
        title = "{kuibit: Analyzing Einstein Toolkit simulations with Python}",
      journal = {The Journal of Open Source Software},
     keywords = {numerical relativity, Python, Einstein Toolkit, astrophysics, Cactus, General Relativity and Quantum Cosmology, Astrophysics - High Energy Astrophysical Phenomena},
         year = 2021,
        month = apr,
       volume = {6},
       number = {60},
          eid = {3099},
        pages = {3099},
          doi = {10.21105/joss.03099},
archivePrefix = {arXiv},
       eprint = {2104.06376},
 primaryClass = {gr-qc},
       adsurl = {https://ui.adsabs.harvard.edu/abs/2021JOSS....6.3099B},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
"""



#!/usr/bin/env python3

# Copyright (C) 2020-2022 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/grid_data.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.uniform_grid` module provides representations of uniform
grids.The object defined is :py:class:`~.UniformGrid`, which represents the
geometry of a uniform Cartesian cell-centered grid.

This is a basic building block of :py:class:`~.UniformGridData`.

"""
import numpy as np


class UniformGrid:
    """Describes the geometry of a regular rectangular dataset, as well as
    information needed to identify the grid if part of refined grid hierarchy
    (namely component number and refinement level). In practice, this a fixed
    refinement level, or part of it (as output by an MPI process).

    This is a standard Cartesian grid that we will describe with the language
    of computer graphics. To make things clear, let's consider a 2D grid (see
    schematics below). We call the lower left corner "origin" or "x0". We call
    the top right corner "x1". The grid is cell-centered (see Fig 2).

    ..code-block::

        Fig 1

         o---------x1
         |          |
         |          |
         |          |
         |          |
        x0----------o

    ..code-block::

         Fig 2, the point sits in the center of a cell.

          --------
          |      |
          |  x0  |
          |      |
          --------

    The concept of ``shape`` is the same as NumPy shape: it's the number of points
    in each dimension. ``dx`` is the spacing (dx, dy, dz, ...). To fully
    describe a grid, one needs the ``origin``, the ``shape``, and ``x1`` or ``dx``.

    (This is the same convention that Carpet has.)

    This class is supposed to be immutable.

    :ivar ~.shape:     Number of points in each dimension.
    :type ~.shape:      1d NumPy arrary or list of int.
    :ivar ~.x0:    Position of cell center with lowest coordinate.
    :type ~.x0:     1d NumPy array or list of float.
    :ivar ~.dx:     If not None, specifies grid spacing, else grid
                      spacing is computed from x0, x1, and shape.
    :type ~.dx:      1d NumPy array or list of float.
    :ivar ~.x1:        If grid spacing is None, this specifies the
                      position of the cell center with largest
                      coordinates.
    :type ~.x1:         1d NumPy array or list of float.
    :ivar ~.ref_level:  Refinement level if this belongs to a hierarchy,
                      else -1.
    :type ~.ref_level:   int
    :ivar ~.component: Component number if this belongs to a hierarchy,
                      else -1.
    :type ~.component:  int
    :ivar ~.num_ghost:    Number of ghost zones (default=0)
    :type ~.num_ghost:     1d NumPy arrary or list of int.
    :ivar ~.time:      Time if that makes sense, else None.
    :type ~.time:       float or None
    :ivar ~.iteration: Iteration if that makes sense, else None.
    :type ~.iteration:  float or None

    """

    def _check_dims(self, var, name):
        """Check that the dimensions are consistent with the shape of the object."""
        if len(var.shape) != 1:
            raise ValueError(
                f"{name} must not be multi-dimensional {var.shape}."
            )
        if len(var) != len(self.shape):
            raise ValueError(
                f"The dimensions of this object are {self.shape.shape}, "
                f"not {var.shape} in {name}."
            )

    def __init__(
        self,
        shape,
        x0,
        x1=None,
        dx=None,
        ref_level=-1,
        component=-1,
        num_ghost=None,
        time=None,
        iteration=None,
    ):
        """
        :param shape:     Number of points in each dimension.
        :type shape:      1d NumPy arrary or list of int.
        :param x0:    Position of cell center with lowest coordinate.
        :type x0:     1d NumPy array or list of float.
        :param dx:     If not None, specifies grid spacing, else grid
                          spacing is computed from x0, x1, and shape.
        :type dx:      1d NumPy array or list of float.
        :param x1:        If grid spacing is None, this specifies the
                          position of the cell center with largest
                          coordinates.
        :type x1:         1d NumPy array or list of float.
        :param ref_level:  Refinement level if this belongs to a hierarchy,
                          else -1.
        :type ref_level:   int
        :param component: Component number if this belongs to a hierarchy,
                          else -1.
        :type component:  int
        :param num_ghost:    Number of ghost zones (default=0)
        :type num_ghost:     1d NumPy arrary or list of int.
        :param time:      Time if that makes sense, else None.
        :type time:       float or None
        :param iteration: Iteration if that makes sense, else None.
        :type iteration:  float or None

        """

        self.__shape = np.atleast_1d(np.array(shape, dtype=int))
        self.__x0 = np.atleast_1d(np.array(x0, dtype=float))

        self._check_dims(self.shape, "shape")
        self._check_dims(self.x0, "x0")

        # Internally, what is important is shape, x0 and dx

        if dx is None:
            if x1 is None:
                raise ValueError("Must provide one between x1 and dx")

            # Here x1 is given, we compute dx. Consider the case
            # with three cells, x0 = (0,0) and x1 = (4,4).
            # Since we work with cell-centered, the x=0 line would
            # look like:
            #
            # --|------|------|------|------|--
            # (0,0)  (1,0)  (2,0)  (3,0)  (4,0)
            #
            # If we want a dx of 1, we have need 5 points. Vice versa,
            # if we have n points, dx is (x1 - x0)/(n - 1).
            x1_arr = np.atleast_1d(x1)
            self._check_dims(x1_arr, "x1")

            if not all(self.x0 <= x1_arr):
                raise ValueError(
                    f"x1 {x1_arr} should be the upper corner (x0 = {x0})"
                )
            # We have to deal with grid with only one point along one direction.
            # When that happens, the only way to produce a grid that makes sense
            # is to have x0 and dx, because for those grids x0 = x1 along the
            # direction that has one single point.
            if np.any(self.shape == 1):
                raise ValueError(
                    "Cannot initialize a grid with a dimension"
                    " that has only one grid point by providing"
                    " x0 and x1. You must provide dx"
                )
            self.__dx = (x1_arr - self.x0) / (self.shape - 1)
        else:
            # Here we assume dx is given, but if also x1 is given, that
            # would may lead to problems if the paramters do not agree. So, we
            # first compute what x1 would be given x0 and dx, then if x1
            # is provided, we compare the result with the given x1. We raise an
            # error if they disagree.
            self.__dx = np.atleast_1d(np.array(dx, dtype=float))
            self._check_dims(self.dx, "dx")
            expected_x1 = self.x0 + (self.shape - 1) * self.dx
            if (x1 is not None) and (
                not np.allclose(expected_x1, x1, atol=1e-14)
            ):
                raise ValueError("Incompatible x1 and dx")

        if num_ghost is None:
            self.__num_ghost = np.zeros_like(self.shape)
        else:
            self.__num_ghost = np.atleast_1d(np.array(num_ghost, dtype=int))
            self._check_dims(self.num_ghost, "num_ghost")

        self.__ref_level = int(ref_level)
        self.__component = int(component)
        self.__time = None if time is None else float(time)
        self.__iteration = None if iteration is None else int(iteration)

        # The coordinates are a widely requested property, so the first time the
        # coordinates 1d are computed, we save them here.
        self.__coordinates_1d = None
        # Same with x1
        self.__x1 = None

        # The method __contains__ is called extremely often when dealing with
        # HierachicalGridData (because it is used to find which subgrid
        # contains a point). So, to make it faster, we save the values of the
        # bottom and upper cell faces (we account for the fact that the grid
        # is cell centered)
        self.__lowest_vertex = None
        self.__highest_vertex = None
        # Same considerations for num_dimensions
        self.__num_dimensions = len(self.shape)

    def __hash__(self):
        """UniformGrid is immutable, we can define an hash as the composition of
        the hashes of the members. This hash is quite slow to compute, so it
        is not useful for caching small computations. Having an hash function
        solidifies the idea that this class is immutable.
        """
        # We convert all the arrays in tuples (because they are hashable)
        hash_shape = hash(tuple(self.shape))
        hash_x0 = hash(tuple(self.x0))
        hash_dx = hash(tuple(self.dx))
        hash_num_ghost = hash(tuple(self.num_ghost))
        hash_ref_level = hash(self.ref_level)
        hash_component = hash(self.component)
        hash_time = hash(self.time)
        hash_iteration = hash(self.iteration)

        # ^ = bitwise xor
        return (
            hash_shape
            ^ hash_x0
            ^ hash_dx
            ^ hash_num_ghost
            ^ hash_ref_level
            ^ hash_component
            ^ hash_time
            ^ hash_iteration
        )

    @property
    def x0(self):
        """Lower corner.

        :returns: Center of lowest corner grid point.
        :rtype: 1d NumPy array
        """
        return self.__x0

    @property
    def shape(self):
        """Number of cells across each dimension.

        :returns: Number of cells across each dimension.
        :rtype: 1d NumPy array
        """
        return self.__shape

    @property
    def x1(self):
        """Upper corner.

        :returns: Center of top corner grid point.
        :rtype: 1d NumPy array
        """
        # We save x1 because it is computed a lot of times
        if self.__x1 is None:
            self.__x1 = self.x0 + (self.shape - 1) * self.dx
        return self.__x1

    @property
    def origin(self):
        """Lower corner.

        Alias for :py:meth:`~.x0`.

        :returns: Center of lowest corner grid point.
        :rtype: 1d NumPy array
        """
        return self.__x0

    @property
    def dx(self):
        """Grid spacing.

        :returns: Cell size across each dimension.
        :rtype: 1d NumPy array
        """
        return self.__dx

    @property
    def delta(self):
        """Grid spacing.

        Alias for :py:meth:`~.dx`.

        :returns: Cell size across each dimension.
        :rtype: 1d NumPy array
        """
        return self.__dx

    @property
    def num_ghost(self):
        """Number of ghost zones.

        :returns: Number of ghost zones across each dimension.
        :rtype: 1d NumPy array
        """
        return self.__num_ghost

    @property
    def ref_level(self):
        """Refinement level number.

        :returns: Refinement level number.
        :rtype: int
        """
        return self.__ref_level

    @property
    def component(self):
        """Component number.

        :returns: Component number.
        :rtype: int
        """
        return self.__component

    @property
    def time(self):
        """Time.

        :returns: Time.
        :rtype: float
        """
        return self.__time

    @property
    def iteration(self):
        """Iteration number

        :returns: Iteration number.
        :rtype: float
        """
        return self.__iteration

    @property
    def dv(self):
        """Volume of a grid cell.

        :returns: Volume of a grid cell.
        :rtype:   float
        """
        return self.dx.prod()

    @property
    def volume(self):
        """Volume of the whole grid.

        :returns: Volume of the whole grid.
        :rtype:   float
        """
        return self.shape.prod() * self.dv

    @property
    def num_dimensions(self):
        """Number of dimensions of the grid.

        :returns: Number of dimensions of the grid.
        :rtype:   float
        """
        return self.__num_dimensions

    @property
    def extended_dimensions(self):
        """Return an array of bools with whether a dimension has more than one
        point or not.

        :returns: Dimensions with more than one point.
        :rtype:   1d NumPy of bools
        """
        return self.shape > 1

    @property
    def num_extended_dimensions(self):
        """Return the number of dimensions with size larger than one gridpoint.

        :returns: The number of extended dimensions (the ones with more than one cell).
        :rtype:   int
        """
        return sum(self.extended_dimensions)

    @property
    def lowest_vertex(self):
        """Return the location of the lowest cell vertex (considering that
        the grid is cell centered).

        :returns: Lowest vertex of the lowest cell.
        :rtype:   1d NumPy array
        """
        if self.__lowest_vertex is None:
            self.__lowest_vertex = self.x0 - 0.5 * self.dx
        return self.__lowest_vertex

    @property
    def highest_vertex(self):
        """Return the location of the highest cell vertex (considering that
        the grid is cell centered).

        :returns: Highest vertex of the highest cell.
        :rtype:   1d NumPy array
        """
        if self.__highest_vertex is None:
            self.__highest_vertex = self.x1 + 0.5 * self.dx
        return self.__highest_vertex

    def indices_to_coordinates(self, indices):
        """Compute coordinate corresponding to one or more grid points.

        :param indices: Grid indices.
        :type indices:  1d array or list of int
        :returns: Corresponding coordinates of the grid points.
        :rtype:   1d NumPy array of float
        """
        # TODO (FEATURE): Add dimensionality checks
        return np.asarray(indices) * self.dx + self.x0

    def coordinates_to_indices(self, coordinates):
        """Find the indices corresponding to the point nearest to the given coordinates.

        :param coordinates: Coordinates.
        :type coordinates:  1d NumPy array or list of float
        :returns: Grid indices of nearest points.
        :rtype:   NumPy array

        """
        # TODO (FEATURE): Add dimensionality checks
        indices = (
            ((np.asarray(coordinates) - self.x0) / self.dx) + 0.5
        ).astype(np.int32)
        return indices

    def __getitem__(self, index):
        """Return the coordinates corresponding to a given (multi-dimensional)
        index.
        """
        index = np.array(index)
        self._check_dims(index, "index")
        coordinate = index * self.dx + self.x0
        if coordinate not in self:
            raise ValueError(f"{index} is not in on the grid")
        return coordinate

    def __contains__(self, point):
        """Test if a coordinate is contained in the grid. The size of the
        grid cells is taken into account, resulting in a cube larger by
        dx/2 on each side compared to the one given by x0, x1.

        :param point: Coordinate to test.
        :type point:  1d NumPy array or list of float.
        :returns:   If point is contained.
        :rtype:     bool
        """

        # A pythonic way to write this function is:
        # if np.any(point < (self.lowest_vertex)) or np.any(
        #     point > (self.highest_vertex)
        # ):
        #     return False
        # return True

        # However, this happens to be not the fastest. This method is called a
        # huge number of times when in HierarchicalGridData methods, because it
        # is the main method to find which grid contains a given point. (method
        # finest_component_at_point). So, it is better to have a less pythonic
        # method that fails as soon as possible.
        for dim in range(self.num_dimensions):
            if not (
                self.lowest_vertex[dim]
                <= point[dim]
                < self.highest_vertex[dim]
            ):
                return False
        return True

    def contains(self, point):
        """Test if a coordinate is contained in the grid. The size of the
        grid cells is taken into account, resulting in a cube larger by
        dx/2 on each side compared to the one given by x0, x1.

        :param point: Coordinate to test.
        :type point:  1d NumPy array or list of float.
        :returns:   If point is contained.
        :rtype:     bool
        """
        return point in self

    @property
    def coordinates_1d(self):
        """Return coordinates of the grid points.

        The return value is a list with the coordinates along each direction.

        :returns: Coordinates of the grid points on each direction.
        :rtype: list of 1d NumPy array

        """
        if self.__coordinates_1d is None:
            self.__coordinates_1d = [
                np.linspace(x0, x1, n)
                for n, x0, x1 in zip(self.shape, self.x0, self.x1)
            ]
        return self.__coordinates_1d

    def coordinates(self, as_meshgrid=False, as_same_shape=False):
        """Return coordinates of the grid points.

        If ``as_meshgrid`` is True, the coordinates are returned as NumPy
        meshgrid. Otherwise, return the coordinates of the grid points as 1D
        arrays (schematically, [array for x coordinates, array for y
        coordinates, ...]).

        If ``as_same_shape`` is True return the coordinates as an array with the
        same shape of self and with values the coordinates. This is useful for
        computations involving the coordinates. The output of ``as_same_shape``
        is the same as using ``np.mgrid``.

        :param as_meshgrid: If True, return the coordinates as meshgrid.
        :type as_meshgrid: bool
        :param as_same_shape: If True, return the coordinates as a list
                              or coordinates with the same shape of self
                              and with values of a given coordinate.
                              For instance, if ``self.num_dimension = 3`` there
                              will be three lists with ``shape = self.shape``.
                              This is equivalent to ``np.mgrid``.
        :type as_same_shape: bool
        :returns:  Grid coordinates.
        :rtype:   list of NumPy arrays with the same shape as grid

        """
        if as_meshgrid and as_same_shape:
            raise ValueError("Cannot ask for both meshgrid and shaped array.")

        if as_meshgrid:
            return np.meshgrid(*self.coordinates_1d)

        if as_same_shape:
            # np.indeces prepares an array in which each element has the value
            # of its index
            indices = np.indices(self.shape)
            # Here we transform the index into coordinate, along each dimension
            return [
                indices[dim] * self.dx[dim] + self.x0[dim]
                for dim in range(0, self.shape.size)
            ]

        return self.coordinates_1d

    def flat_dimensions_removed(self):
        """Return a new :py:class:`~.UniformGrid` with dimensions that have no
        flat dimensions (dimensions with only one grid point).

        :returns: Return a new grid without flat dimensions.
        :rtype: :py:class:`~.UniformGrid`
        """

        # We need this infrastructure to slice UniformGridData

        copied = self.copy()

        # We have to save this, otherwise it would be recomputed at every line,
        # affecting the result.
        extended_dims = copied.extended_dimensions

        copied.__shape = copied.__shape[extended_dims]
        copied.__x0 = copied.__x0[extended_dims]
        copied.__dx = copied.__dx[extended_dims]
        copied.__num_ghost = copied.__num_ghost[extended_dims]
        return copied

    def ghost_zones_removed(self):
        """Return a new :py:class:`~.UniformGrid` with ghostzones removed.

        :returns: Return a new grid without ghost zones.
        :rtype: :py:class:`~.UniformGrid`
        """
        copied = self.copy()

        # We remove twice the number of ghost zones because there are
        # lower and upper ghostzones
        copied.__shape = self.shape - 2 * self.num_ghost
        # We "push x0 inside the grid"
        copied.__x0 = copied.__x0 + self.num_ghost * self.dx
        copied.__num_ghost = np.zeros_like(copied.shape)
        return copied

    def shifted(self, shift):
        """Return a new UniformGrid with coordinates shifted by the given amount.

        ``x -> x + shift``.

        :param shift: Amount to shift coordinates.
        :type shift: 1d NumPy array
        :returns: New grid with coordinates shifted.
        :rtype: :py:class:`~.UniformGrid`

        """
        shift = np.asarray(shift)
        self._check_dims(shift, "shift")

        copied = self.copy()

        # We only need to shift x0 because x1 is computed from x0 using dx
        copied.__x0 = copied.__x0 + shift
        return copied

    def copy(self):
        """Return a deep copy.

        :returns:  Deep copy of the UniformGrid.
        :rtype:    :py:class:`~.UniformGrid`
        """
        return type(self)(
            self.shape,
            self.x0,
            dx=self.dx,
            x1=self.x1,
            ref_level=self.ref_level,
            component=self.component,
            num_ghost=self.num_ghost,
            time=self.time,
            iteration=self.iteration,
        )

    def __eq__(self, other):
        """Test if two :py:class:`~.UniformGrid` are the equal up to numerical
        precision.
        """
        if not isinstance(other, type(self)):
            return False
        # Time and iterations can be None, so we check them independently
        if self.time is not None and other.time is not None:
            time_bool = np.isclose(self.time, other.time, atol=1e-14)
        else:
            # In this case one of the two is None (or both)
            time_bool = self.time is other.time

        if self.iteration is not None and other.iteration is not None:
            iteration_bool = np.isclose(
                self.iteration, other.iteration, atol=1e-14
            )
        else:
            # In this case one of the two is None (or both)
            iteration_bool = self.iteration is other.iteration

        return (
            np.array_equal(self.shape, other.shape)
            and np.allclose(self.x0, other.x0, atol=1e-14)
            and np.allclose(self.dx, other.dx, atol=1e-14)
            and np.allclose(self.num_ghost, other.num_ghost, atol=1e-14)
            and np.allclose(self.ref_level, other.ref_level, atol=1e-14)
            and np.allclose(self.component, other.component, atol=1e-14)
            and time_bool
            and iteration_bool
        )

    def __str__(self):
        """:returns: a string describing the geometry."""
        return f"""Shape            = {self.shape}
Num ghost zones  = {self.num_ghost}
Ref. level       = {self.ref_level}
Component        = {self.component}
x0               = {self.x0}
x0/dx            = {self.x0/self.dx}
x1               = {self.x1}
x1/dx            = {self.x1/self.dx}
Volume           = {self.volume}
dx               = {self.dx}
Time             = {self.time}
Iteration        = {self.iteration}
"""
